(* Bare language - Definition v3 pages 8,9,70,71 *)
(* modified to have ident in place of con and var *)

(*
$File: Common/DecGrammar.sml $
$Date: 1992/04/07 14:54:27 $
$Revision: 1.18 $
$Locker:  $
*)

(*$DecGrammar: LAB SCON CON TYVAR TYCON EXCON STRID DEC_GRAMMAR*)

functor DecGrammar(structure GrammarInfo: sig  type GrammarInfo  
                                               type PostInfo
                                               val emptyInfo : PostInfo 
                                          end

		   structure Lab   : LAB
		   structure SCon  : SCON
		   structure Con   : CON
		   structure TyVar : TYVAR
		   structure TyCon : TYCON
		   structure Excon : EXCON
		   structure StrId : STRID
(* PVADD *)
(* Haven't added any lines here but have update signature *)
(* to sig eqtype longid end instead of sig type longid end *)
(* Similarly, have made id and eqtype as well *)
(* This is OK, cos they are in their respective structures *)
		    
		   structure Ident : sig  eqtype longid  end
		   structure Id    : sig  eqtype id  end
		  ): DEC_GRAMMAR =
struct
  type lab    = Lab.lab      (* labels *)
   and scon   = SCon.scon    (* special constants *)
   and con    = Con.con      (* constructors *)
   and id     = Id.id        (* identifiers - variables or constructors *)
   and longid = Ident.longid (* long identifiers - variables or constructors *)
   and tyvar  = TyVar.SyntaxTyVar   (* type variables *)
   and tycon  = TyCon.tycon  (* type constructors *)
   and longtycon = TyCon.longtycon  (* long type constructors *)
   and excon     = Excon.excon      (* exception constructors *)
   and longstrid = StrId.longstrid  (* structure identifiers *)
(* PVADD *)
(**) and strid = StrId.strid (* structure identifier *)
(* need this to do equality on longstrids *)


  type info = GrammarInfo.GrammarInfo

  type PostInfo = GrammarInfo.PostInfo

  val emptyInfo = GrammarInfo.emptyInfo

  datatype 'a op_opt = OP_OPT of 'a * bool
  datatype 'a WithInfo = WITH_INFO of info * 'a

  datatype atexp =
	SCONatexp of info * scon |         
	IDENTatexp of info * longid op_opt |
	RECORDatexp of info * exprow Option |
	LETatexp of info * dec * exp |
	PARatexp of info * exp

  and opid = OPID of longid * bool

  and exprow =
	EXPROW of info * lab * exp * exprow Option

  and exp =
(*EML*) EXISTSexp of info * match |
(*EML*) FORALLexp of info * match |
(*EML*) TERMINATESexp of info * atexp |
	ATEXPexp of info * atexp |
	APPexp of info * exp * atexp |
	TYPEDexp of info * exp * ty |
	HANDLEexp of info * exp * match |
        RAISEexp of info * exp |
	FNexp of info * match |
	UNRES_INFIXexp of info * atexp list

  and match =
        MATCH of info * mrule * match Option

  and mrule =
        MRULE of info * pat * exp

  and dec =  
(*EML*) EQTYPEdec of info * typbind |
(*EML*) SPANTYPEdec of info * datbind |
(*EML*) AXIOMdec of info * ax |
	VALdec of info * valbind |
	UNRES_FUNdec of info * FValBind |
	TYPEdec of info * typbind |
	DATATYPEdec of info * datbind |
	ABSTYPEdec of info * datbind * dec |
	EXCEPTIONdec of info * exbind |
	LOCALdec of info * dec * dec |
	OPENdec of info * longstrid WithInfo list |
	SEQdec of info * dec * dec |
	INFIXdec of info * int Option * id list |
	INFIXRdec of info * int Option * id list |
	NONFIXdec of info * id list |
	EMPTYdec of info

  and valbind =
(*EML*) UNDEFvalbind of info * pat * (*?*) valbind Option |
	PLAINvalbind of info * pat * exp * valbind Option |
	RECvalbind of info * valbind

  and FValBind = FVALBIND of info * FClause * FValBind Option
  and FClause = FCLAUSE of info * atpat list * ty Option * exp * FClause Option |
                UNDEFFClause of info * atpat list * ty Option * ty Option * FClause Option
  and typbind =
(*EML*) UNDEFtypbind of info * tyvar list * tycon * (*?*) typbind Option |
        TYPBIND of info * tyvar list * tycon * ty * typbind Option

  and datbind =
        DATBIND of info * tyvar list * tycon * conbind * datbind Option

  and conbind =
        CONBIND of info * con op_opt * ty Option * conbind Option

  and exbind =
        EXBIND of info * excon op_opt * ty Option * exbind Option |
        EXEQUAL of info * excon op_opt * longid op_opt * exbind Option

  and atpat =
        WILDCARDatpat of info |
	SCONatpat of info * scon |
	LONGIDatpat of info * longid op_opt |
	RECORDatpat of info * patrow Option |
	PARatpat of info * pat

  and patrow =
        DOTDOTDOT of info |
        PATROW of info * lab * pat * patrow Option

  and pat =
        ATPATpat of info * atpat |
        CONSpat of info * longid op_opt * atpat |
        TYPEDpat of info * pat * ty |
        LAYEREDpat of info * id op_opt * ty Option * pat |
	UNRES_INFIXpat of info * atpat list

  and ty =
        TYVARty of info * tyvar |
        RECORDty of info * tyrow Option |
        CONty of info * ty list * longtycon |
        FNty of info * ty * ty |
        PARty of info * ty

  and tyrow =
        TYROW of info * lab * ty * tyrow Option

(*EML*)and ax = AXIOMS of info * exp * ax Option




(* PVADD *)
(**) 
(**) open Lab
(**) open SCon
(**) open Con
(**) open TyVar
(**) open TyCon
(**) open Excon
(**) open StrId
(**) open Ident
(**) open Id
(**)
(* All following functions added *)
(**)
(**)  fun eq_option _ (None,None) =
(**)    true
(**)  |   eq_option eq_f (Some(a),Some(b)) =
(**)    eq_f(a,b)
(**)  |   eq_option _ _ =
(**)    false
(**)
(**)  fun eq_list eq_f ([],[]) =
(**)    true
(**)  |   eq_list eq_f ((x::xs),(y::ys)) =
(**)    (eq_f(x,y)) andalso (eq_list eq_f (xs,ys))
(**)  |   eq_list _ (_,_) =
(**)    false
(**)
(**)  fun eq_int(i1,i2) =
(**)    i1=i2
(**)
(**)  fun eq_string(s1,s2) =
(**)    s1=s2
(**)
(**)  fun eq_real(r1,r2) =
(**)    r1=r2
(**)
(**)  fun eq_lab(lab1,lab2) =
(**)    lab1=lab2
(**)
(**)  fun eq_scon(INTEGER(int1),INTEGER(int2)) =
(**)    eq_int(int1,int2)
(**)  |   eq_scon(STRING(string1),STRING(string2)) =
(**)    eq_string(string1,string2)
(**)  |   eq_scon(REAL(real1),REAL(real2)) =
(**)    eq_real(real1,real2)
(**)  |   eq_scon(_,_) =
(**)    false
(**)
(**)  fun eq_con(c1,c2) =
(**)    c1=c2
(**)
(**)  fun eq_id(i1,i2) =
(**)    i1=i2
(**)
(**)  fun eq_longid(l1,l2) =
(**)    l1=l2
(**)
(**)
(**)  fun eq_tyvar(t1,t2) =
(**)    t1=t2
(**)
(**)  fun eq_tycon(t1,t2) =
(**)    t1=t2
(**)
(**)  fun eq_longtycon(l1,l2) =
(**)    l1=l2
(**)
(**)  fun eq_excon(e1,e2) =
(**)    e1=e2
(**)
(**)  fun eq_strid(s1,s2) =
(**)    s1=s2
(**)
(**)  fun eq_longstrid(longstrid1,longstrid2) =
(**)    let
(**)      val (stridlist1,strid1) =
(**)        explode_longstrid longstrid1
(**)      val (stridlist2,strid2) =
(**)        explode_longstrid longstrid2
(**)    in
(**)      (eq_list eq_strid (stridlist1,stridlist2)) andalso
(**)      (eq_strid(strid1,strid2))
(**)    end
(**)
(**)  fun eq_op_opt eq_f ((OP_OPT(a,tfa)),(OP_OPT(b,tfb))) =
(**)    (tfa=tfb) andalso (eq_f(a,b))
(**)
(**)  fun eq_with_info eq_f ((WITH_INFO(_,a)),(WITH_INFO(_,b))) =
(**)    eq_f(a,b)
(**)
(**)  fun eq_atexp(SCONatexp(_,scon1),SCONatexp(_,scon2)) =
(**)    eq_scon(scon1,scon2)
(**)  |   eq_atexp(IDENTatexp(_,longid_op_opt1),
(**)               IDENTatexp(_,longid_op_opt2)) =
(**)    eq_op_opt eq_longid (longid_op_opt1,longid_op_opt2)
(**)  |   eq_atexp(RECORDatexp(_,exprow_option1),RECORDatexp(_,exprow_option2)) =
(**)    eq_option eq_exprow (exprow_option1,exprow_option2)
(**)  |   eq_atexp(LETatexp(_,dec1,exp1),LETatexp(_,dec2,exp2)) =
(**)    (eq_dec(dec1,dec2)) andalso (eq_exp(exp1,exp2))
(**)  |   eq_atexp(PARatexp(_,exp1),PARatexp(_,exp2)) =
(**)    eq_exp(exp1,exp2)
(**)  |   eq_atexp(_,_) =
(**)    false
(**)
(**)  and eq_exprow(EXPROW(_,lab1,exp1,exprow_option1),EXPROW(_,lab2,exp2,exprow_option2)) =
(**)    (eq_lab(lab1,lab2)) andalso
(**)    (eq_exp(exp1,exp2)) andalso
(**)    (eq_option eq_exprow (exprow_option1,exprow_option2))
(**)  
(**)  and eq_exp(EXISTSexp(_,match1),EXISTSexp(_,match2)) =
(**)    eq_match(match1,match2)
(**)  |   eq_exp(FORALLexp(_,match1),FORALLexp(_,match2)) =
(**)    eq_match(match1,match2)
(**)  |   eq_exp(TERMINATESexp(_,atexp1),TERMINATESexp(_,atexp2)) =
(**)    eq_atexp(atexp1,atexp2)
(**)  |   eq_exp(ATEXPexp(_,atexp1),ATEXPexp(_,atexp2)) =
(**)    eq_atexp(atexp1,atexp2)
(**)  |   eq_exp(APPexp(_,exp1,atexp1),APPexp(_,exp2,atexp2)) =
(**)    (eq_exp(exp1,exp2)) andalso (eq_atexp(atexp1,atexp2))
(**)  |   eq_exp(TYPEDexp(_,exp1,ty1),TYPEDexp(_,exp2,ty2)) =
(**)    (eq_exp(exp1,exp2)) andalso (eq_ty(ty1,ty2))
(**)  |   eq_exp(HANDLEexp(_,exp1,match1),HANDLEexp(_,exp2,match2)) =
(**)    (eq_exp(exp1,exp2)) andalso (eq_match(match1,match2))
(**)  |   eq_exp(RAISEexp(_,exp1),RAISEexp(_,exp2)) =
(**)    eq_exp(exp1,exp2)
(**)  |   eq_exp(FNexp(_,match1),FNexp(_,match2)) =
(**)    eq_match(match1,match2)
(**)  |   eq_exp(UNRES_INFIXexp(_,atexplist1),UNRES_INFIXexp(_,atexplist2)) =
(**)    eq_list eq_atexp (atexplist1,atexplist2)
(**)  |   eq_exp(_,_) =
(**)    false
(**)
(**)  and eq_match(MATCH(_,mrule1,match_option1),MATCH(_,mrule2,match_option2)) =
(**)    (eq_mrule(mrule1,mrule2)) andalso 
(**)    (eq_option eq_match (match_option1,match_option2))
(**)
(**)  and eq_mrule(MRULE(_,pat1,exp1),MRULE(_,pat2,exp2)) =
(**)    (eq_pat(pat1,pat2)) andalso (eq_exp(exp1,exp2))
(**)
(**)  and eq_dec(EQTYPEdec(_,typbind1),EQTYPEdec(_,typbind2)) =
(**)    eq_typbind(typbind1,typbind2)
(**)  |   eq_dec(SPANTYPEdec(_,datbind1),SPANTYPEdec(_,datbind2)) =
(**)    eq_datbind(datbind1,datbind2)
(**)  |   eq_dec(AXIOMdec(_,ax1),AXIOMdec(_,ax2)) =
(**)    eq_ax(ax1,ax2)
(**)  |   eq_dec(VALdec(_,valbind1),VALdec(_,valbind2)) =
(**)    eq_valbind(valbind1,valbind2)
(**)  |   eq_dec(UNRES_FUNdec(_,FValBind1),UNRES_FUNdec(_,FValBind2)) =
(**)    eq_fvalbind(FValBind1,FValBind2)
(**)  |   eq_dec(TYPEdec(_,typbind1),TYPEdec(_,typbind2)) =
(**)    eq_typbind(typbind1,typbind2)
(**)  |   eq_dec(DATATYPEdec(_,datbind1),DATATYPEdec(_,datbind2)) =
(**)    eq_datbind(datbind1,datbind2)
(**)  |   eq_dec(ABSTYPEdec(_,datbind1,dec1),ABSTYPEdec(_,datbind2,dec2)) =
(**)    (eq_dec(dec1,dec2)) andalso (eq_datbind(datbind1,datbind2))
(**)  |   eq_dec(EXCEPTIONdec(_,exbind1),EXCEPTIONdec(_,exbind2)) =
(**)    eq_exbind(exbind1,exbind2)
(**)  |   eq_dec(LOCALdec(_,dec1,dec3),LOCALdec(_,dec2,dec4)) =
(**)    (eq_dec(dec1,dec2)) andalso (eq_dec(dec3,dec4))
(**)  |   eq_dec(OPENdec(_,longstridwithinfolist1),OPENdec(_,longstridwithinfolist2)) =
(**)    (eq_list (eq_with_info eq_longstrid) 
(**)             (longstridwithinfolist1,longstridwithinfolist2))
(**)  |   eq_dec(SEQdec(_,dec1,dec3),SEQdec(_,dec2,dec4)) =
(**)    (eq_dec(dec1,dec2)) andalso (eq_dec(dec3,dec4))
(**)  |   eq_dec(INFIXdec(_,intoption1,idlist1),INFIXdec(_,intoption2,idlist2)) =
(**)    (eq_option eq_int (intoption1,intoption2)) andalso
(**)    (eq_list eq_id (idlist1,idlist2))
(**)  |   eq_dec(INFIXRdec(_,intoption1,idlist1),INFIXRdec(_,intoption2,idlist2)) =
(**)    (eq_option eq_int (intoption1,intoption2)) andalso
(**)    (eq_list eq_id (idlist1,idlist2))
(**)  |   eq_dec(NONFIXdec(_,idlist1),NONFIXdec(_,idlist2)) =
(**)    (eq_list eq_id (idlist1,idlist2))
(**)  |   eq_dec(EMPTYdec(_),EMPTYdec(_)) =
(**)    true
(**)  |   eq_dec(_,_) =
(**)    false
(**)
(**)  and eq_valbind(UNDEFvalbind(_,pat1,valbindoption1),
(**)                 UNDEFvalbind(_,pat2,valbindoption2)) =
(**)    (eq_pat(pat1,pat2)) andalso 
(**)    (eq_option eq_valbind (valbindoption1,valbindoption2))
(**)  |   eq_valbind(PLAINvalbind(_,pat1,exp1,valbindoption1),
(**)                 PLAINvalbind(_,pat2,exp2,valbindoption2)) =
(**)    (eq_pat(pat1,pat2)) andalso 
(**)    (eq_exp(exp1,exp2)) andalso
(**)    (eq_option eq_valbind (valbindoption1,valbindoption2))
(**)  |   eq_valbind(RECvalbind(_,valbind1),RECvalbind(_,valbind2)) =
(**)    eq_valbind(valbind1,valbind2)
(**)  |   eq_valbind(_,_) =
(**)    false
(**)
(**)  and eq_fvalbind(FVALBIND(_,FClause1,FValBindoption1),
(**)                  FVALBIND(_,FClause2,FValBindoption2)) =
(**)    (eq_fclause(FClause1,FClause2)) andalso
(**)    (eq_option eq_fvalbind (FValBindoption1,FValBindoption2))
(**)
(**)  and eq_fclause(FCLAUSE(_,atpatlist1,tyoption1,exp1,FClauseoption1),
(**)                 FCLAUSE(_,atpatlist2,tyoption2,exp2,FClauseoption2)) =
(**)    (eq_list eq_atpat (atpatlist1,atpatlist2)) andalso
(**)    (eq_option eq_ty (tyoption1,tyoption2)) andalso
(**)    (eq_exp (exp1,exp2)) andalso
(**)    (eq_option eq_fclause (FClauseoption1,FClauseoption2))
(**)  |   eq_fclause(UNDEFFClause(_,atpatlist1,tyoption1,tyoption3,FClauseoption1),
(**)                 UNDEFFClause(_,atpatlist2,tyoption2,tyoption4,FClauseoption2)) =
(**)    (eq_list eq_atpat (atpatlist1,atpatlist2)) andalso
(**)    (eq_option eq_ty (tyoption1,tyoption2)) andalso
(**)    (eq_option eq_ty (tyoption3,tyoption4)) andalso
(**)    (eq_option eq_fclause (FClauseoption1,FClauseoption2))
(**)  |   eq_fclause(_,_) =
(**)    false
(**)
(**)  and eq_typbind(UNDEFtypbind(_,tyvarlist1,tycon1,typbindoption1),
(**)                 UNDEFtypbind(_,tyvarlist2,tycon2,typbindoption2)) =
(**)    (eq_list eq_tyvar (tyvarlist1,tyvarlist2)) andalso
(**)    (eq_tycon (tycon1,tycon2)) andalso
(**)    (eq_option eq_typbind (typbindoption1,typbindoption2))
(**)  |   eq_typbind(TYPBIND(_,tyvarlist1,tycon1,ty1,typbindoption1),
(**)                 TYPBIND(_,tyvarlist2,tycon2,ty2,typbindoption2)) =
(**)    (eq_list eq_tyvar (tyvarlist1,tyvarlist2)) andalso
(**)    (eq_tycon (tycon1,tycon2)) andalso
(**)    (eq_ty (ty1,ty2)) andalso
(**)    (eq_option eq_typbind (typbindoption1,typbindoption2))
(**)  |   eq_typbind(_,_) =
(**)    false
(**)
(**)  and eq_datbind(DATBIND(_,tyvarlist1,tycon1,conbind1,datbindoption1),
(**)                 DATBIND(_,tyvarlist2,tycon2,conbind2,datbindoption2)) =
(**)    (eq_list eq_tyvar (tyvarlist1,tyvarlist2)) andalso
(**)    (eq_tycon(tycon1,tycon2)) andalso
(**)    (eq_conbind(conbind1,conbind2)) andalso
(**)    (eq_option eq_datbind (datbindoption1,datbindoption2))
(**)
(**)  and eq_conbind(CONBIND(_,conopopt1,tyoption1,conbindoption1),
(**)                 CONBIND(_,conopopt2,tyoption2,conbindoption2)) =
(**)    (eq_op_opt eq_con (conopopt1,conopopt2)) andalso
(**)    (eq_option eq_ty (tyoption1,tyoption2)) andalso
(**)    (eq_option eq_conbind (conbindoption1,conbindoption2))
(**)
(**)  and eq_exbind(EXBIND(_,exconopopt1,tyoption1,exbindoption1),
(**)                EXBIND(_,exconopopt2,tyoption2,exbindoption2)) =
(**)    (eq_op_opt eq_excon (exconopopt1,exconopopt2)) andalso
(**)    (eq_option eq_ty (tyoption1,tyoption2)) andalso
(**)    (eq_option eq_exbind (exbindoption1,exbindoption2))
(**)  |   eq_exbind(EXEQUAL(_,exconopopt1,longidopopt1,exbindoption1),
(**)                EXEQUAL(_,exconopopt2,longidopopt2,exbindoption2)) =
(**)    (eq_op_opt eq_excon (exconopopt1,exconopopt2)) andalso
(**)    (eq_op_opt eq_longid (longidopopt1,longidopopt2)) andalso
(**)    (eq_option eq_exbind (exbindoption1,exbindoption2))
(**)  |   eq_exbind(_,_) =
(**)    false
(**)
(**)  and eq_atpat(WILDCARDatpat(_),WILDCARDatpat(_)) =
(**)    true
(**)  |   eq_atpat(SCONatpat(_,scon1),SCONatpat(_,scon2)) =
(**)    eq_scon(scon1,scon2)
(**)  |   eq_atpat(LONGIDatpat(_,longidopopt1),LONGIDatpat(_,longidopopt2)) =
(**)    eq_op_opt eq_longid (longidopopt1,longidopopt2)
(**)  |   eq_atpat(RECORDatpat(_,patrowoption1),RECORDatpat(_,patrowoption2)) =
(**)    eq_option eq_patrow (patrowoption1,patrowoption2)
(**)  |   eq_atpat(PARatpat(_,pat1),PARatpat(_,pat2)) =
(**)    eq_pat(pat1,pat2)
(**)  |   eq_atpat(_,_) =
(**)    false
(**)
(**)  and eq_patrow(DOTDOTDOT(_),DOTDOTDOT(_)) =
(**)    true
(**)  |   eq_patrow(PATROW(_,lab1,pat1,patrowoption1),
(**)                PATROW(_,lab2,pat2,patrowoption2)) =
(**)    (eq_lab (lab1,lab2)) andalso
(**)    (eq_pat (pat1,pat2)) andalso
(**)    (eq_option eq_patrow (patrowoption1,patrowoption2))
(**)  |   eq_patrow(_,_) =
(**)    false
(**)
(**)  and eq_pat(ATPATpat(_,atpat1),ATPATpat(_,atpat2)) =
(**)    eq_atpat(atpat1,atpat2)
(**)  |   eq_pat(CONSpat(_,longidopopt1,atpat1),CONSpat(_,longidopopt2,atpat2)) =
(**)    (eq_op_opt eq_longid (longidopopt1,longidopopt2)) andalso
(**)    (eq_atpat (atpat1,atpat2))
(**)  |   eq_pat(TYPEDpat(_,pat1,ty1),TYPEDpat(_,pat2,ty2)) =
(**)    (eq_pat(pat1,pat2)) andalso
(**)    (eq_ty(ty1,ty2))
(**)  |   eq_pat(LAYEREDpat(_,idopopt1,tyoption1,pat1),
(**)             LAYEREDpat(_,idopopt2,tyoption2,pat2)) =
(**)    (eq_op_opt eq_id (idopopt1,idopopt2)) andalso
(**)    (eq_option eq_ty (tyoption1,tyoption2)) andalso
(**)    (eq_pat(pat1,pat2))
(**)  |   eq_pat(UNRES_INFIXpat(_,atpatlist1),UNRES_INFIXpat(_,atpatlist2)) =
(**)    eq_list eq_atpat (atpatlist1,atpatlist2)
(**)  |   eq_pat(_,_) =
(**)    false
(**)
(**)  and eq_ty(TYVARty(_,tyvar1),TYVARty(_,tyvar2)) =
(**)    eq_tyvar(tyvar1,tyvar2)
(**)  |   eq_ty(RECORDty(_,tyrowoption1),RECORDty(_,tyrowoption2)) =
(**)    eq_option eq_tyrow (tyrowoption1,tyrowoption2)
(**)  |   eq_ty(CONty(_,tylist1,longtycon1),CONty(_,tylist2,longtycon2)) =
(**)    (eq_list eq_ty (tylist1,tylist2)) andalso
(**)    (eq_longtycon(longtycon1,longtycon2))
(**)  |   eq_ty(FNty(_,ty1,ty3),FNty(_,ty2,ty4)) =
(**)    (eq_ty(ty1,ty2)) andalso (eq_ty(ty3,ty4))
(**)  |   eq_ty(PARty(_,ty1),PARty(_,ty2)) =
(**)    eq_ty (ty1,ty2)
(**)  |   eq_ty(_,_) =
(**)    false
(**)
(**)  and eq_tyrow(TYROW(_,lab1,ty1,tyrowoption1),
(**)               TYROW(_,lab2,ty2,tyrowoption2)) =
(**)    (eq_lab(lab1,lab2)) andalso (eq_ty(ty1,ty2)) andalso
(**)    (eq_option eq_tyrow (tyrowoption2,tyrowoption2))
(**)
(**)  and eq_ax(AXIOMS(_,exp1,axoption1),AXIOMS(_,exp2,axoption2)) =
(**)    (eq_exp(exp1,exp2)) andalso (eq_option eq_ax (axoption1,axoption2))

end;
