(* Elaborator for Core Language Declarations*)

(*
$File: Common/ElabDec.sml $
$Date: 92/07/01 18:29:50 $
$Revision: 1.62 $
$Locker:  $
*)

(*$ElabDec:
	DEC_GRAMMAR ENVIRONMENTS STATOBJECT CON LAB VAR ELABDEC SOURCE_INFO
	ERROR_INFO TYPE_INFO GRAMMAR_INFO IDENT PPDECGRAMMAR REPORT PRETTYPRINT
	FLAGS CRASH
 *)

functor ElabDec(structure IG: DEC_GRAMMAR

		structure OG: DEC_GRAMMAR
		  sharing type IG.lab       = OG.lab
		      and type IG.con       = OG.con
		      and type IG.scon      = OG.scon
		      and type IG.tyvar     = OG.tyvar
		      and type IG.tycon     = OG.tycon
		      and type IG.longtycon = OG.longtycon
		      and type IG.longstrid = OG.longstrid
		      and type IG.excon     = OG.excon

	       (* resolved identifiers *)
		structure R:
		  sig
		    eqtype longvar
		       and longcon
		       and longexcon

		    datatype longid = LONGVAR of longvar
				    | LONGCON of longcon
				    | LONGEXCON of longexcon
		  end
		  sharing type R.longid = OG.longid

		structure Environments: ENVIRONMENTS
		  sharing type Environments.longid    = IG.longid
		      and type Environments.longtycon = IG.longtycon
		      and type Environments.longstrid = IG.longstrid
		      and type Environments.tycon     = IG.tycon
		      and type Environments.valbind   = IG.valbind
		      and type Environments.excon     = IG.excon
		      and type Environments.con       = IG.con
		      and type Environments.id        = IG.id

		structure StatObject: STATOBJECT
		  sharing type StatObject.TypeScheme   = Environments.TypeScheme
		      and type StatObject.SyntaxTyVar  = IG.tyvar
		      and type StatObject.longid       = IG.longid
		      and type StatObject.longcon      = R.longcon
		      and type StatObject.longvar      = R.longvar
		      and type StatObject.longexcon    = R.longexcon
		      and type StatObject.TypeFcn      = Environments.TypeFcn
		      and type StatObject.Substitution
			       = Environments.Substitution
		      and type StatObject.Type   = Environments.Type
		      and type StatObject.TyVar  = Environments.TyVar
		      and type StatObject.TyName = Environments.TyName
                      and type StatObject.TyNameSet 
                               = Environments.TyNameSet
		      and type StatObject.tycon = IG.tycon
		      and type StatObject.scon  = IG.scon
		      and type StatObject.lab   = IG.lab
		      and type StatObject.id    = IG.id

	       (* Unqualified identifiers are always converted into vars *)
		  sharing type StatObject.var = OG.id

	       (* Con needed for compiler info. *)
		structure Con: CON
		  sharing type Con.longid = IG.longid
		      and type StatObject.longcon = Con.longcon
		      and type Environments.con = Con.con

	       (* Lab needed for the equality attribute. *)
		structure Lab: LAB
		  sharing type OG.lab = Lab.lab

	       (* We need Var to map vars (PostElabDecGrammar.id) into ids
		  (PreElabDecGrammar.id) for the RECvalbind analysis. *)

		structure Var: VAR
		  sharing type IG.id = Var.id
		      and type OG.id = Var.var

		structure SourceInfo: SOURCE_INFO

		structure ErrorInfo: ERROR_INFO
		  sharing type ErrorInfo.Type  = StatObject.Type
		      and type ErrorInfo.TyVar = StatObject.TyVar
		      and type ErrorInfo.longid = StatObject.longid
		      and type ErrorInfo.longstrid = Environments.longstrid
		      and type ErrorInfo.longtycon = Environments.longtycon

		structure TypeInfo: TYPE_INFO
		  sharing type TypeInfo.lab     = IG.lab
		      and type TypeInfo.longcon = R.longcon

		structure GrammarInfo: GRAMMAR_INFO
		  sharing type GrammarInfo.ErrorInfo = ErrorInfo.info
		      and type GrammarInfo.TypeInfo = TypeInfo.info
		      and type GrammarInfo.PreElabGrammarInfo = IG.info
		      and type GrammarInfo.PostElabGrammarInfo = OG.info

		structure PPInDecGrammar: PPDECGRAMMAR
		  sharing PPInDecGrammar.G = IG

	        structure Report: REPORT

		structure PP: PRETTYPRINT
		  sharing type StatObject.StringTree
			       = Environments.StringTree
			       = PPInDecGrammar.StringTree
			       = PP.StringTree
		      and type PP.Report = Report.Report

		structure Flags: FLAGS
		structure Crash: CRASH
	       ) : ELABDEC =
  struct
    type PreElabGrammarInfo  = GrammarInfo.PreElabGrammarInfo
     and PostElabGrammarInfo = GrammarInfo.PostElabGrammarInfo

    val okConv = GrammarInfo.convertGrammarInfo
				(* Convert pre-elab. info into post-elab, with
				   empty Error and Type fields. *)

    fun errorConv(i: PreElabGrammarInfo, e: ErrorInfo.info)
          : PostElabGrammarInfo =
      GrammarInfo.addPostElabErrorInfo (okConv i) e
				(* Convert pre-elab. to post-elab. adding
				   an error tag at the same time. The assumption
				   (forced by this type conversion) is that
				   there can only be one error tag per info. *)

    fun lookupIdError(i: PreElabGrammarInfo, longid: StatObject.longid)
          : PostElabGrammarInfo =
      errorConv(i, ErrorInfo.LOOKUP_LONGID longid)

    fun lookupTyConError(i: PreElabGrammarInfo,
			 longtycon: Environments.longtycon
			): PostElabGrammarInfo =
      errorConv(i, ErrorInfo.LOOKUP_LONGTYCON longtycon)

    infixr on
    val op on = StatObject.on

    infixr onC
    val op onC = Environments.onC

    infixr onVE
    val op onVE = Environments.onVE

    infixr oo
    val op oo = StatObject.oo

    infixr C_cplus_E
    val op C_cplus_E = Environments.C_cplus_E

    infixr C_cplus_TE
    val op C_cplus_TE = Environments.C_cplus_TE

    infixr C_cplus_VE_and_TE
    val op C_cplus_VE_and_TE = Environments.C_cplus_VE_and_TE

    type PreElabDec = IG.dec and PostElabDec = OG.dec
    type PreElabTy  = IG.ty  and PostElabTy  = OG.ty
(*EML*)
    type PreElabAx  = IG.ax  and PostElabAx  = OG.ax

    type Type    = StatObject.Type
    type Env     = Environments.Env
    type Context = Environments.Context
    type Substitution = StatObject.Substitution

    fun pr(msg: string, t: PP.StringTree): unit =
      Report.print(Report.decorate(msg, PP.reportStringTree t))

    fun where list elem =
      case List.index (General.curry (op =) elem) list
	of OK n => n
	 | Fail _ => Crash.impossible "ElabDec.where"

    (* Hooks needed by the compiler:

	  o for a RECORDatpat, a type-info node of type LAB_INFO in each
	    PATROW.

	  o for a CONSpat or constructor LONGIDatpat or constructor IDENTatexp,
	    a type-info node of type CON_INFO.

	  o for a exception constructor IDENTatexp, a type-info node of
	    type EXCON_INFO.
     *)

    local
      open GrammarInfo TypeInfo
    in
     (* MEMO: no duplication checks here (or anywhere else!) *)
      fun addTypeInfo_CON(postElabInfo, C, isArrow, id) =
	let
	  val con = #2(Con.decompose(StatObject.mk_longcon id))
	  val cons = Environments.lookupFellowCons(C, id)
	in
	  addPostElabTypeInfo postElabInfo
	    (CON_INFO{numCons=List.size cons,
		      index=where cons con,
		      functional=isArrow
		     }
	    )
	end

      fun addTypeInfo_EXCON(postElabInfo, tau) =
	(* The excon carries a value if the type tau is functional,
	   and doesn't if it's `exn'. *)
	addPostElabTypeInfo postElabInfo
			    (EXCON_INFO{functional=StatObject.isTypeArrow tau})

      fun addTypeInfo_LAB(postElabInfo, index) =
	addPostElabTypeInfo postElabInfo (LAB_INFO{index=index})
    end

    (********
    Get a fresh type variable
    *********
    Returns a type variable with the imperative and equality attributes set to be false.
    ********)

    fun freshTyVar() =
      StatObject.freshTyVar{equality=false, imperative=false}

    fun freshType() = StatObject.mkTypeTyVar(freshTyVar())

   (* When we get elaboration errors we often need to return a bogus type.
      It's best if that's just 'a. *)

    val bogus_Type = freshType

    fun Unify(tau, tau', i): Substitution * PostElabGrammarInfo =
      case StatObject.unify(tau, tau')
	of Some subst => (subst, okConv i)
         | None =>
	     (StatObject.bogus_Subst,
	      errorConv(i, ErrorInfo.UNIFICATION(tau, tau'))
	     )

    (********
     Update the Error information if a typbind does not deny eqality in T:
     Extended ML only.
     ********)

    fun denies_equality_error(info) =
      GrammarInfo.addPostElabErrorInfo info ErrorInfo.DENIES_EQUALITY

    (********
     Update the Error information if a "eqtype" declaration does 
     not admit equality : Extended ML only.
    ********)

    fun admits_equality_error(info) =
      GrammarInfo.addPostElabErrorInfo info ErrorInfo.ADMITS_EQUALITY

    (********
     Update the Error information if "forall" or "exists" do not have
     correct type : Extended ML only.
    ********)

    fun predicate_error(info, tau) =
      GrammarInfo.addPostElabErrorInfo info (ErrorInfo.PREDICATE(tau))

    (********
     Update the Error information if expressions in axioms do not have
     correct type : Extended ML only.
    ********)

    fun logexp_error(info,tau) =
      GrammarInfo.addPostElabErrorInfo info (ErrorInfo.LOGEXP(tau))


   (* Traversal of patterns to build a VE. We only do this for `rec'
      bindings. Arguably, we should reject records and so on right here,
      but the caller isn't set up to deal with that (and ignoring them
      would give unbound var errors for actual occuring vars), so we work for
      all patterns. We reject illegal pattern forms in the `rec' post-pass. *)

   (*MEMO: these should all be returning sets. *)

    fun dom_pat(C: Environments.Context,pat: IG.pat): Environments.id list =
      let
	fun dom_patrow(patrow: IG.patrow): Environments.id list =
	  case patrow
	    of IG.DOTDOTDOT _ =>
	         nil

	     | IG.PATROW(_, _, pat, None) =>
		 dom_pat' pat

	     | IG.PATROW(_, _, pat, Some patrow') =>
		 dom_pat' pat @ dom_patrow patrow'

	and dom_atpat(atpat: IG.atpat): Environments.id list =
	  case atpat
	    of IG.WILDCARDatpat _ => nil
	     | IG.SCONatpat(_, _) => nil
	     | IG.RECORDatpat(_, None) => nil
	     | IG.RECORDatpat(_, Some patrow) => dom_patrow patrow
	     | IG.PARatpat(_, pat) => dom_pat' pat

	     | IG.LONGIDatpat(i, IG.OP_OPT(longid, _)) =>
		 (case Environments.Lookup_longid(C, longid)
		    of Some(Environments.LONGCON _) => nil
		     | Some(Environments.LONGEXCON _) => nil
		     | _ => case StatObject.decomposeLongId longid
			      of (nil, id) => [id]
			       | _ => Crash.impossible "ElabDec.dom_atpat"
		 )

	and dom_pat'(pat: IG.pat): Environments.id list =
	  case pat
	    of IG.ATPATpat(_, atpat) => dom_atpat atpat
	     | IG.CONSpat(_, _, atpat) => dom_atpat atpat
	     | IG.TYPEDpat(_, pat, _) => dom_pat' pat
	     | IG.LAYEREDpat(_, IG.OP_OPT(id, _), _, pat) => id :: dom_pat' pat
	     | IG.UNRES_INFIXpat _ => Crash.impossible "dom_pat'(UNRES_INFIX)"
      in
	dom_pat' pat
      end

    (********
    Find the variables bound in a valbind
    *********
    There is no checking for multiply declared variables
    ********)

    fun dom_vb(C: Environments.Context, vb: IG.valbind): Environments.id list =
      case vb
	of IG.PLAINvalbind(_, pat, _, vb_opt) =>
	     dom_pat(C, pat) @ (case vb_opt
				  of Some vb => dom_vb(C, vb)
				   | None => nil
			       )

         | IG.RECvalbind(_, vb) =>
	     dom_vb(C, vb)

         | IG.UNDEFvalbind(_, pat, vb_opt) => 
             dom_pat(C,pat) @ (case vb_opt
				  of Some vb => dom_vb(C, vb)
				   | None => nil
			       )

    (*********
    Find the TE to be used initially in elaborating a datbind
    **********
    We determine the correct equality attributes when we maximise equality
    *********)

    local
      fun make_TE tyvar_list tycon =
      let
	val arity =
	  List.size tyvar_list

	val tyname =
	  StatObject.freshTyName {name = tycon, arity = arity, equality = false}

	val typeFcn =
	  StatObject.TyName_in_TypeFcn tyname

	val tystr =
	  Environments.mkTyStr(typeFcn, Environments.emptyCE)
      in
	Environments.singleTE(tycon, tystr)
      end
    in
      fun initial_TE (IG.DATBIND(_, tyvar_list, tycon, _, None)) =
	  make_TE tyvar_list tycon
	| initial_TE (IG.DATBIND(_, tyvar_list, tycon, _, Some datbind)) =
	  Environments.TE_plus_TE(make_TE tyvar_list tycon, initial_TE datbind)
    end

   (********
   Test for Correct typing of axioms 
   ********
   Extended ML functions for booleans 
   ********
   A bad solution - but we do not know what the type name of boolean is and
   so must look it up in the environment 
   ********)

   fun boolean(TE) = 
     case Environments.boolean(TE) of
          Some(tystr) => 
            let
              val (theta,CE) = Environments.unTyStr tystr
              val boolName = 
                  case (StatObject.unTyName_TypeFcn theta) of 
                       None => Crash.impossible "ElabDec.booleans" 
                     | Some t => t
              val boolCons = StatObject.mkConsType([],boolName)
              val boolTy   = StatObject.mkTypeConsType boolCons
            in
              boolTy
            end
        | None      => Crash.impossible "ElabDec.booleans"

   fun TyEquals(tau,tau') = 
    let 
      val subst = StatObject.unify(tau,tau')
    in 
      (* Two types are equal if there unifier is empty *)

       case subst of 
            None   => true
          | Some S => StatObject.emptySubst S
     end
      
   fun is_predicate(TE,tau) = 
        let
          val tau'    = StatObject.unTypeArrow(tau)
          val boolean = boolean(TE)
        in  
          case tau' of 
               None            => false
             | Some((ty1,ty2)) => TyEquals(boolean,ty2)
        end
 
    fun is_bool(TE,ty) = 
     let
       val boolean = boolean(TE)
     in
       TyEquals(ty,boolean)
     end


   (********
   The test for "denying equality" - Extended ML (version 2)
   *********)

   fun denies_equality(C : Environments.Context, theta : Environments.TypeFcn) =
      (*  The "denies" check for "?" typbind: ensures that the new type 
          name for "?" will not admit equality in a (possibly) recursive 
          typbind  *)
      let
        val tyname = StatObject.unTyName_TypeFcn(theta)
      in
        case tyname of
            None    => true
          | Some(t) => 
                if StatObject.isIn(t, Environments.T_of_C(C)) then
                   true
                else
                   not(StatObject.admits_equality(theta))
		   (* know that theta = \a.t where t is a type name *)
      end

   (**************************************************************************)

   (* addLabelInfo: given a RecType and a PATROW, populate the info fields
      with LAB_INFO specifying which label index (numbered canonically from
      0) is associated with each label. This is needed by the compiler. *)

    fun addLabelInfo(recType, patrow) =
      let
	val sortedLabs = StatObject.sortedLabsOfRecType recType

	fun f(OG.PATROW(i, lab, pat, patrow_opt)) =
	      let
		val index = where sortedLabs lab
	      in
		OG.PATROW(addTypeInfo_LAB(i, index),
			  lab, pat,
			  case patrow_opt
			    of Some patrow => Some(f patrow)
			     | None => None
			 )
	      end

	  | f(OG.DOTDOTDOT i) = OG.DOTDOTDOT i
      in
	f patrow
      end

    (********************************************************)
    (*      Elaboration (type checking)                     *)
    (********************************************************)


    (****** atomic expressions - Definition v3 page 23 ******)

    fun elab_atexp (C : Environments.Context, atexp : IG.atexp) :
	(Substitution * StatObject.Type * OG.atexp) =

	case atexp of

	  (* special constants *)
	  IG.SCONatexp(i, scon) =>
	    (StatObject.Id, StatObject.GetTypescon scon, OG.SCONatexp(okConv i,scon))

	  (* identifiers - variables or constructors *)     
	| IG.IDENTatexp(i, IG.OP_OPT(ident, withOp)) =>
	    (case Environments.Lookup_longid(C,ident) of

	      (* Variable *)
	      Some(Environments.LONGVAR sigma) =>
		 (StatObject.Id, StatObject.instance sigma,
		  OG.IDENTatexp(
		     okConv i,
		     OG.OP_OPT(R.LONGVAR(StatObject.mk_longvar ident), withOp)
		  )
		 )

	      (* Constructor *)
	    | Some(Environments.LONGCON sigma) =>
		let
		  val tau = StatObject.instance sigma
		in
		  (StatObject.Id, tau,
		   let
		     val longcon = StatObject.mk_longcon ident
		   in
		     OG.IDENTatexp(addTypeInfo_CON(okConv i, C,
						   StatObject.isTypeArrow tau,
						   ident
						  ),
				   OG.OP_OPT(R.LONGCON longcon, withOp)
				  )
		   end
		  )
		end

	     (* Exception constructor *)
	   | Some(Environments.LONGEXCON tau) =>
		(StatObject.Id,
		 tau,
		 let
		   val excon = StatObject.mk_longexcon ident
		 in
		   OG.IDENTatexp(addTypeInfo_EXCON(okConv i, tau),
				 OG.OP_OPT(R.LONGEXCON excon, withOp)
				)
		 end)

	     (* Not found in current context *)
	   | None =>
	       (StatObject.bogus_Subst, bogus_Type(),
		 OG.IDENTatexp(lookupIdError(i, ident),
			       OG.OP_OPT(R.LONGVAR StatObject.bogusVar, withOp)
			      )
	       )
	  )

	  (* record expression *)
	| IG.RECORDatexp(i, None) =>
	    (StatObject.Id, StatObject.TypeUnit, OG.RECORDatexp(okConv i,None)) 

	  (* record expression *)
	| IG.RECORDatexp(i, Some exprow) =>
	    let
	      val (S, rho, out_exprow) = elab_exprow(C,exprow)
	    in
	      (S, StatObject.mkTypeRecType rho,
	       OG.RECORDatexp(okConv i,Some out_exprow)) 
	    end 

	  (* let expression *)
	| IG.LETatexp(i, dec, exp) => 
	    let
	      val (S1, E, out_dec)   = elab_dec(C,dec)
	      val (S2, tau, out_exp) = elab_exp((S1 onC C) C_cplus_E E, exp)
	    in
	      (S2 oo S1, tau, OG.LETatexp(okConv i,out_dec,out_exp))
	    end

	  (* parenthesised expression *)
	| IG.PARatexp(i, exp) =>
	    let val (S, tau, out_exp) = elab_exp(C,exp)
	    in (S, tau, OG.PARatexp(okConv i,out_exp)) end

    (******** expression rows - Definition v3 page 24 ********)

    and elab_exprow (C : Environments.Context, exprow : IG.exprow) :
	(Substitution * StatObject.RecType * OG.exprow) =

	case exprow of 

	  (* Expression row *)
	  IG.EXPROW(i, lab, exp, None) =>
	    let
	      val (S, tau, out_exp) = elab_exp(C, exp)
	      val rho = StatObject.addField (lab,tau) StatObject.emptyRecType
	    in
	      (S, rho, OG.EXPROW(okConv i,lab,out_exp,None))
	    end

	  (* Expression row *)
	| IG.EXPROW(i, lab, exp, Some exprow) =>
	    let
	      val (S1, tau, out_exp   ) = elab_exp(C, exp)
	      val (S2, rho, out_exprow) = elab_exprow(S1 onC C,exprow)
	      val rho1 = StatObject.addField (lab,S2 on tau) rho
	    in
	      (S2 oo S1, rho1, OG.EXPROW(okConv i,lab,out_exp,Some out_exprow))
	    end

    (******** expressions - Definition v3 page 24 ********)

    and elab_exp(C : Environments.Context, exp : IG.exp) : 
	(Substitution * StatObject.Type * OG.exp) =
      let
        val _ =
	  if Flags.DEBUG_ELABDEC then
	    pr("elab_exp: ", PPInDecGrammar.layoutExp exp)
	  else ()

	val (S, ty, exp') = elab_exp'(C, exp)

	val _ =
	  if Flags.DEBUG_ELABDEC then
	    pr("giving:   ", StatObject.layoutType ty)
	  else ()
      in
	(S, ty, exp')
      end

    and elab_exp'(C, exp) =
      case exp

	   (* Atomic expression *)
	of IG.ATEXPexp(i, atexp) =>
	     let
	       val (S, tau, out_atexp) = elab_atexp(C, atexp)
	     in
	       (S, tau, OG.ATEXPexp(okConv i, out_atexp))
	     end

	   (* Application expression *)
	 | IG.APPexp(i, exp, atexp) => 
	     let
	       val (S1, tau1, out_exp)   = elab_exp(C, exp)
	       val (S2, tau2, out_atexp) = elab_atexp(S1 onC C, atexp)
	       val new   = freshType()
	       val arrow = StatObject.mkTypeArrow(tau2,new) 
	       val (S3, i') = Unify(arrow, S2 on tau1, i)
	     in
	       (S3 oo S2 oo S1, S3 on new, OG.APPexp(i', out_exp, out_atexp))
	     end

	   (* Typed expression *)
	 | IG.TYPEDexp(i, exp, ty) =>
	     let
	       val (S1, tau, out_exp) = elab_exp(C, exp)
	       val (tau', out_ty) = elab_ty(S1 onC C, ty)
	       val (S2, i') = Unify(tau, tau', i)
	     in
	       (S2 oo S1, S2 on tau', OG.TYPEDexp(i', out_exp, out_ty))
	     end

	   (* Handle exception *)
	 | IG.HANDLEexp(i, exp, match) =>
	     let
	       val (S1, tau1, out_exp)   = elab_exp(C, exp)
	       val (S2, tau2, out_match) = elab_match(S1 onC C, match)
	       val matchTy = StatObject.mkTypeArrow(StatObject.TypeExn, tau1)
	       val (S3, i') = Unify(matchTy, tau2, i)
	     in
	       (S3 oo S2 oo S1, (S3 oo S2) on tau1,
		OG.HANDLEexp(i', out_exp, out_match))
	     end

	   (* Raise exception *)
	 | IG.RAISEexp(i, exp) =>
	     let
	       val (S1, tau1, out_exp) = elab_exp(C, exp)
	       val exnType = StatObject.TypeExn
	       val (S2, i')    = Unify(exnType, S1 on tau1, i)
	     in
	       (S2 oo S1, freshType(), OG.RAISEexp(i', out_exp))
	     end

	   (* Function expression *)
	 | IG.FNexp(i, match) => 
	     let
	       val (S, tau, out_match) = elab_match(C, match)
	     in
	       (S, tau, OG.FNexp(okConv i, out_match))
	     end
 
         (*  Extended ML: "Syntax, Semantics and Dynamics Semantics of 
             Extended ML (version 2)" *)

         (* Forall Quantification - rule 14.2 *)

         | IG.FORALLexp(i,match) => 
             let
               val (S1, tau, out_match) = elab_match(C, match)
               val  TE = Environments.TE_of_C(C)
             in
               if is_predicate(TE,tau) then
                 let
                   (*  Guarded by the "is_predicate" test  *)
                   val boolean = boolean(TE)
                 in
                   (S1, boolean, OG.FORALLexp(okConv i,out_match))
                 end
               else
                 let 
                   val i' = predicate_error(okConv i,tau)
                 in  
                   (S1, tau, OG.FORALLexp(i',out_match))
                 end 
             end

         (* Exists Quantification - rule 14.1 *)

         | IG.EXISTSexp(i,match) => 
             let
               val (S1, tau, out_match) = elab_match(C, match)
               val  TE = Environments.TE_of_C(C)
             in
               if is_predicate(TE, tau) then
                 let
                   (*  Guarded by the "is_predicate" test  *)
                   val boolean = boolean(TE)
                 in
                   (S1, boolean, OG.EXISTSexp(okConv i,out_match))
                 end
               else
                 let 
                   val i' = predicate_error(okConv i,tau)
                 in  
                   (S1, tau, OG.EXISTSexp(i', out_match))
                 end 
             end
         
         (* Divergence predicate - rule 14.3 *)

         | IG.TERMINATESexp(i,atexp) => 
             let
               val (S, tau, out_exp) = elab_atexp(C, atexp)
               val TE = Environments.TE_of_C(C)
               val boolStr = Environments.boolean(TE)
               (* Only Tempory since we may have rebound
                  "bool" to something else in TE  *)
             in
               case boolStr of 
                  Some(_) =>
                     (S, boolean(TE), OG.TERMINATESexp(okConv i,out_exp))
                | None =>(* Tempory *)
                     (StatObject.Id, freshType(), OG.TERMINATESexp(okConv i,out_exp)) 
             end

	 | IG.UNRES_INFIXexp _ =>
	     Crash.impossible "elab_exp(UNRES_INFIX)"


    (******** matches - Definition v3 page 25 ********)

    and elab_match (C : Environments.Context, match : IG.match) :
	(Substitution * StatObject.Type * OG.match) =

	case match of

	  (* Match *)
	  IG.MATCH(i, mrule, None) =>
	    let val (S, tau, out_mrule) = elab_mrule(C,mrule)
	    in (S, tau, OG.MATCH(okConv i,out_mrule,None)) end

	  (* Match *)
	| IG.MATCH(i, mrule, Some match') =>
	    let
	      val (S ,tau ,out_mrule) = elab_mrule(C,mrule)
	      val (S',tau',out_match) = elab_match(S onC C,match')
	      val (S'', i') = Unify(tau',S' on tau,i)
	    in
	      (S'' oo S' oo S, S'' on tau',
	       OG.MATCH(i', out_mrule, Some out_match))
	    end

    (******** match rules - Definition v3 page 25 ********)

    and elab_mrule (C : Environments.Context, mrule : IG.mrule) : 
	(Substitution * StatObject.Type * OG.mrule) =

	case mrule of

	  (* Match rule *)
	  IG.MRULE(i, pat, exp) =>
	    let
	      val ((VE,tau),out_pat) = elab_pat(C,pat)
	      val (S,tau',  out_exp) = elab_exp(Environments.C_plus_VE(C,VE),exp)
	    in
	      (S, StatObject.mkTypeArrow(S on tau,tau'),
	       OG.MRULE(okConv i,out_pat,out_exp))
	    end

    (******** declarations - Definition v3 page 25 ********)

    and elab_dec(C: Environments.Context, dec: IG.dec) :
	(Substitution * Environments.Env * OG.dec) =

	case dec of

	   (* Value declaration *)
	   IG.VALdec(i, valbind) =>
	     let
	       val U =
		 Environments.Scoped_TyVars(valbind, Environments.U_of_C C)

	       val (S, VE, out_valbind) =
		 elab_valbind(Environments.C_plus_U(C,U), valbind)

	       val VE' =
		 Environments.Clos(S onC C, valbind, VE)

	       val i' =
		 case Environments.intersect(Environments.tyvarsVE VE', U)
		   of nil => okConv i
		    | tyvars => errorConv(i, ErrorInfo.FREE_TYVARS tyvars)
	     in
	       (S, Environments.VE_in_E VE', OG.VALdec(i', out_valbind)) 
	     end

	   (* `fun'-declaration *)
	 | IG.UNRES_FUNdec _ => Crash.impossible "elab_dec(UNRES_FUN)"

	   (* Datatype declaration *)
	 | IG.DATATYPEdec(i, datbind) =>
	     let
	       val TE = initial_TE datbind
	       val ((VE1, TE1), out_datbind) = elab_datbind(C C_cplus_TE TE, datbind)
	       val (VE2, TE2) = Environments.maximise_equality (VE1, TE1)
	     in
	       (StatObject.Id, Environments.VE_and_TE_in_E(VE2, TE2),
		OG.DATATYPEdec(okConv i, out_datbind))
	     end

	   (* Abstype declaration *)
	 | IG.ABSTYPEdec(i, datbind, dec) =>
	     let
	       val TE = initial_TE datbind
	       val ((VE1, TE1), out_datbind) = elab_datbind(C C_cplus_TE TE, datbind)
	       val (VE2, TE2) = Environments.maximise_equality (VE1, TE1)
	       val (S, E, out_dec) = elab_dec(C C_cplus_VE_and_TE (VE2,TE2), dec)
	     in
	       (S, Environments.ABS(TE2, E),
		OG.ABSTYPEdec(okConv i, out_datbind, out_dec))
	     end

	   (* Exception declaration *)
	 | IG.EXCEPTIONdec(i, exbind) =>
	     let
	       val (EE, out_exbind) = elab_exbind(C, exbind)
	       val VE = Environments.VE_of_EE EE
	     in
	       (StatObject.Id,
		Environments.VE_and_EE_in_E(VE, EE),
		OG.EXCEPTIONdec(okConv i, out_exbind))
	     end

	   (* Local declaration *)
	 | IG.LOCALdec(i, dec1, dec2) =>
	     let
	       val (S1, E1, out_dec1) = elab_dec(C,dec1)
	       val (S2, E2, out_dec2) = elab_dec((S1 onC C) C_cplus_E E1,dec2)
	     in
	       (S2 oo S1, E2, OG.LOCALdec(okConv i,out_dec1,out_dec2))
	     end

	   (* Open declaration *)
	 | IG.OPENdec(i, list) =>
	     let
	       fun process(E0, list)
		   : Env * Environments.longstrid OG.WithInfo list =
		 case list
		   of IG.WITH_INFO(i, longstrid) :: rest =>
		        (case Environments.Lookup_longstrid(C, longstrid)
			   of Some Str =>
			        let
				  val (_, E) = Environments.unStr Str
				  val (E', rest') = process(E0, rest)
				in
				  (Environments.E_plus_E(E, E'),
				   OG.WITH_INFO(okConv i, longstrid) :: rest'
				  )
				end

			    | None =>	(* Lookup failure: process rest of
					   list and build env regardless. *)
				let
				  val (E', rest') = process(E0, rest)
				  val ei = ErrorInfo.LOOKUP_LONGSTRID longstrid
				in
				  (E', OG.WITH_INFO(errorConv(i, ei), longstrid)
				       :: rest'
				  )
				end
			)

		    | nil => (Environments.emptyE, nil)

	       val (E', list') = process(Environments.emptyE, list)
	     in
	       (StatObject.Id, E', OG.OPENdec(okConv i, list'))
	     end

	 | IG.INFIXdec(i, prec, ids) =>
	     (StatObject.Id,
	      Environments.VE_in_E Environments.emptyVE,
	      OG.INFIXdec(okConv i,
			  case prec
			    of Some p => Some p
			     | None => None,
			  map StatObject.mk_var ids
			 )
	     )

	 | IG.INFIXRdec(i, prec, ids) =>
	     (StatObject.Id,
	      Environments.VE_in_E Environments.emptyVE,
	      OG.INFIXRdec(okConv i,
			   case prec
			     of Some p => Some p
			      | None => None,
			   map StatObject.mk_var ids
			  )
	     )

	 | IG.NONFIXdec(i, ids) =>
	     (StatObject.Id,
	      Environments.VE_in_E Environments.emptyVE,
	      OG.NONFIXdec(okConv i, map StatObject.mk_var ids)
	     )

	   (* Empty declaration *)
	 | IG.EMPTYdec(i) =>
	     (StatObject.Id, Environments.VE_in_E Environments.emptyVE, OG.EMPTYdec(okConv i))

	   (* Sequential declaration *)
	 | IG.SEQdec(i, dec1, dec2) =>
	     let
	       val (S1, E1, out_dec1) = elab_dec(C,dec1)
	       val (S2, E2, out_dec2) = elab_dec((S1 onC C) C_cplus_E E1,dec2)
	       val E1' = Environments.onE(S2,E1)
	     in
	       (S2 oo S1,
		Environments.E_plus_E(E1',E2),
		OG.SEQdec(okConv i,out_dec1,out_dec2)) 
	     end


         (*  Extended ML: "Syntax, Semantics and Dynamics Semantics of 
             Extended ML (version 2)" *)

           (* Type declaration - rule 18 *)

         | IG.TYPEdec(i, typbind) =>
	    let
               fun conjunct x y = x andalso y (* ... because "andalso" cannot
                                                  be passed as a function 
                                                  argument to "map" in nj-sml *)

               (*  proj : TyStr -> TyFcn *)
               fun proj(TyStr) = 
                 let 
                   val (TyFcn,CE) = Environments.unTyStr TyStr 
                 in  
                   TyFcn 
                 end

	       (* Note that no substitutions are produced *)
	       val (TE, out_typbind) = elab_typbind(C, typbind)

               val rangeTE = List.map proj (Environments.TErange TE)

               (* Now a list of truth values: one for each TypeFcn in rangeTE *)
               val denyList = List.map (fn x => denies_equality(C,x)) rangeTE

               (*  ... and finally the "denies equality" check  *)
               val TE_denies_equality = List.foldL conjunct true denyList
	     in
               if TE_denies_equality then 
	          (StatObject.Id, Environments.TE_in_E(TE), OG.TYPEdec(okConv i, out_typbind))
               else
                  let 
                    val i' = denies_equality_error(okConv i)
                  in  
                    (StatObject.Id, Environments.TE_in_E(TE), 
                     OG.TYPEdec(i', out_typbind))
                  end
	     end

	 (* Eqtype declaration - rule 18.5 *)

       | IG.EQTYPEdec(i, typbind) =>
	   let             
             fun proj(TyStr) = 
               let 
                 val (Theta,CE) = Environments.unTyStr TyStr
               in
                 Theta
               end
             
             fun conjunct x y = x andalso y (* ... because "andalso" cannot
                                                be passed as a function 
                                                argument to "map" in nj-sml *)

	     (* Note that no substitutions are produced *)
	     val (TE, out_typbind) = elab_typbind(C, typbind)

             (* For all Type Functions, theta in TE, theta admits equality *)
             val rangeTE  = List.map proj (Environments.TErange TE)
             val rangeTE' = List.map StatObject.admits_equality rangeTE
             val admits_equality =
                 List.foldL conjunct true rangeTE'
	   in
	     if admits_equality then
                (StatObject.Id, Environments.TE_in_E TE, OG.EQTYPEdec(okConv i, out_typbind))
             else 
                let 
                  val i' = admits_equality_error(okConv i)
                in  
                  (StatObject.Id, Environments.TE_in_E(TE), 
                   OG.EQTYPEdec(i', out_typbind))
                end
	   end

         (* Spantype declarartion - rule 19.5 *)

       | IG.SPANTYPEdec(i, datbind) =>
           let
             val TE = initial_TE datbind
             val ((VE1, TE1), out_datbind) = 
                              elab_datbind(C C_cplus_TE TE, datbind)
             val TE2 = Environments.Abst TE1 (* same as Abs(TE) *)
           in
             (StatObject.Id, Environments.VE_and_TE_in_E(VE1,TE2),
                             OG.SPANTYPEdec(okConv i, out_datbind))
           end

         (* Axiom declaration - rule 25.5 *)

       | IG.AXIOMdec(i, ax) =>
           let
             val (E, out_axiom) = elab_axiom(C,ax)
           in
             (StatObject.Id, Environments.emptyE, 
              OG.AXIOMdec(okConv i, out_axiom))
           end

    (****** value bindings - Definition v3 page 26 ******)

    and elab_valbind(C: Environments.Context, valbind: IG.valbind)
          : (Substitution * Environments.VarEnv * OG.valbind) =

	case valbind of

	(* Simple value binding *)
        IG.PLAINvalbind(i, pat, exp, valbind_opt) =>
	  let
	    val ((VE,tau), out_pat) = elab_pat(C, pat)
	    val (S1, tau1, out_exp) = elab_exp(C, exp)
	    val (S2, i') = Unify(S1 on tau, tau1, i)

	    val (S3, VE', valbind_opt') =
	      case valbind_opt
		of Some valbind =>
		     let
		       val (S, VE, vb) =
			 elab_valbind((S2 oo S1) onC C, valbind)
		     in
		       (S, VE, Some vb)
		     end

		 | None =>
		     (StatObject.Id, Environments.emptyVE, None)
	  in 
	    (S3 oo S2 oo S1, 
	     Environments.VE_plus_VE((S3 oo S2 oo S1) onVE VE, VE'),
	     OG.PLAINvalbind(i', out_pat, out_exp, valbind_opt')
	    )
	  end

				(* Recursive value binding. Rather tricky
				   because we have to plant error info after
				   the second pass. Make that `very tricky.' *)
      | IG.RECvalbind(i, valbind) => 
	  let
				(* Function to unify the occurrence of a
				   variable in two VE's. The result is a
				   substitution and an ErrorInfo tag. *)
	    fun processID(i, VE, VE', var): Substitution * PostElabGrammarInfo =
	      case
		(Environments.lookupVE(VE, Var.unVar var),
		 Environments.lookupVE(VE', Var.unVar var))
	      of
		(Some(Environments.LONGVAR sigma1),
		 Some(Environments.LONGVAR sigma2)) =>
		  let
		    val (_, tau1) = StatObject.unTypeScheme sigma1
		    val (_, tau2) = StatObject.unTypeScheme sigma2
		  in
		    case StatObject.unify(tau1, tau2) of
		      Some S =>
			(S, i)
		    | None =>
			(StatObject.bogus_Subst,
			 GrammarInfo.addPostElabErrorInfo i
			 (ErrorInfo.UNIFICATION(tau1, tau2)))
		  end

	       | _ => Crash.impossible "ElabDec.unifyVE"

				(* Traverse the out_valbind, doing a
				   unification (and adding ErrorInfo if reqd.)
				   and giving a subst. at each stage. The
				   ErrorInfo goes into the pattern...
				   ...somewhere... *)
	    fun traverseRecValbind(VE, VE', vb): Substitution * OG.valbind =
	      case vb
		of OG.PLAINvalbind(i, pat, exp, vb_opt) =>
		     let
		       val (S, pat') = traverseRecPat(VE, VE', pat)

		       val (S', vb_opt') =
			 case vb_opt
			   of Some vb =>
			        let
				  val (S', vb') =
				    traverseRecValbind
				    (S onVE VE, S onVE VE', vb)
				in
				  (S' oo S, Some vb')
				end

			    | None => (S, None)
		     in
		       (S', OG.PLAINvalbind(i, pat', exp, vb_opt'))
		     end

		 | OG.RECvalbind(i, vb) =>
		     let
		       val (S, vb') = traverseRecValbind(VE, VE', vb)
		     in
		       (S, OG.RECvalbind(i, vb'))
		     end
 
                 | OG.UNDEFvalbind(i, pat, vb_opt) =>
		     let
		       val (S, pat') = traverseRecPat(VE, VE', pat)

		       val (S', vb_opt') =
			 case vb_opt
			   of Some vb =>
			        let
				  val (S', vb') =
				    traverseRecValbind
				    (S onVE VE, S onVE VE', vb)
				in
				  (S' oo S, Some vb')
				end

			    | None => (S, None)
		     in
		       (S', OG.UNDEFvalbind(i, pat', vb_opt'))
		     end

	    and traverseRecPat(VE, VE', pat): Substitution * OG.pat =
	      case pat
		of OG.ATPATpat(i, atpat) =>
		     let
		       val (S, atpat') = traverseRecAtpat(VE, VE', atpat)
		     in
		       (S, OG.ATPATpat(i, atpat'))
		     end

				(* We can reject a CONS node. *)
		 | OG.CONSpat(i, id, atpat) =>
		     (StatObject.bogus_Subst,
		      OG.CONSpat(
		        GrammarInfo.addPostElabErrorInfo
			  i ErrorInfo.NOTALLOWEDINREC,
			id, atpat
		      )
		     )

		 | OG.TYPEDpat(i, pat, ty) =>
		     let
		       val (S, pat') = traverseRecPat(VE, VE', pat)
		     in
		       (S, OG.TYPEDpat(i, pat', ty))
		     end

		 | OG.LAYEREDpat(i, id as OG.OP_OPT(id', withOp),
				 ty_opt, pat
				) =>
		     let
		       val (S, pat') = traverseRecPat(VE, VE', pat)
		       val (S', i') = processID(i, S onVE VE, S onVE VE', id')
		     in
		       (S oo S', OG.LAYEREDpat(i', id, ty_opt, pat'))
		     end
	    
	         | OG.UNRES_INFIXpat _ =>
		     Crash.impossible "traverseRecPat(UNRES_INFIX)"

	    and traverseRecAtpat(VE, VE', atpat): Substitution * OG.atpat =
	      case atpat
		of OG.WILDCARDatpat _ =>
		     (StatObject.Id, atpat)

		 | OG.SCONatpat _ =>
		     (StatObject.Id, atpat)

		 | OG.LONGIDatpat(i, longid as OG.OP_OPT(longid', withOp)) =>
		     (case longid'
			of R.LONGVAR longvar =>
			     (case StatObject.decomposeLongVar longvar
			        of (nil, var) =>
				     let
				       val (S, i') = processID(i, VE, VE', var)
				     in
				       (S, OG.LONGIDatpat(i', longid))
				     end
				    
			         | _ =>
				     Crash.impossible "traverseRecAtpat(longid)"
			     )

	                 | _ =>		(* LONGCON, LONGEXCON allowed. *)
			     (StatObject.Id, atpat)
		     )

		 | OG.RECORDatpat(i, patrowOpt) =>
		     (StatObject.bogus_Subst,
		      OG.RECORDatpat(
			GrammarInfo.addPostElabErrorInfo
			  i ErrorInfo.NOTALLOWEDINREC,
			patrowOpt
		      )
		     )

		 | OG.PARatpat(i, pat) =>
		     let
		       val (S, pat') = traverseRecPat(VE, VE', pat)
		     in
		       (S, OG.PARatpat(i, pat'))
		     end

				(* set up a value environment, VE,
				   for the recursively declared values *)

	    val domain_list = dom_vb(C, valbind)

	    fun freshTypeScheme() =
	      StatObject.Type_in_TypeScheme(freshType())

	    fun setup id VE =
	      Environments.VE_plus_VE(
		VE,
		Environments.singleVarVE(id, freshTypeScheme())
	      )

	    val VE = List.foldL setup Environments.emptyVE domain_list
				(* VE now maps each rec identifier to 'a. *)

				(* Proceed with type checking. The ErrorInfo
				   tags for the rec identifiers will be
				   untouched (I hope, since we might assign
				   them further on). *)
	    val (S, VE', valbind') =
	      elab_valbind(Environments.C_plus_VE(C, VE), valbind)

				(* Post-pass, to patch up the rec identifiers
				   and plant unification error tags: *)
	    val (S', valbind'') =
	      traverseRecValbind(S onVE VE, VE', valbind')

	    val VE'' = S' onVE VE'

	    val _ =
	      if Flags.DEBUG_ELABDEC then
		let
		  val t = PP.NODE{start="{", finish="}", indent=0,
				  children=[Environments.layoutVE VE''],
				  childsep=PP.NONE
				 }
		in
		  pr("RECvalbind: ", t)
		end
	      else ()
	  in
	    (S' oo S, VE'', OG.RECvalbind(okConv i, valbind''))
	  end


      (*  Extended ML: "Syntax, Semantics and Dynamics Semantics of 
          Extended ML (version 2)" *)

        (* undefined valbind - rule 27.5 *)

      | IG.UNDEFvalbind(i,pat,None) =>
          let 
            val ((VE,tau), out_pat) = elab_pat(C, pat)
          in
            (StatObject.Id, VE, OG.UNDEFvalbind(okConv i,out_pat,None))
          end

      | IG.UNDEFvalbind(i,pat,Some(valbind))  =>
          let
            val ((VE,tau), out_pat) = elab_pat(C,pat)
            val (S1, VE', out_valbind) = elab_valbind(C, valbind)
          in
            (S1, Environments.VE_plus_VE((S1 onVE VE),VE'), 
                 OG.UNDEFvalbind(okConv i,out_pat,Some(out_valbind)))
          end


    (******* type bindings - Definition v3 page 27 *******)

    and elab_typbind (C : Environments.Context, typbind : IG.typbind) :
      (Environments.TyEnv * OG.typbind) =

      case typbind of

	(* Type binding *)
	IG.TYPBIND(i, tyvar_list, tycon, ty, typbind_opt) =>
	  let
	    val TyVar_list =
	      map (fn tv => StatObject.mkExplicitTyVar tv) tyvar_list
	    val (tau, out_ty) = elab_ty(C, ty)
	    val typeFcn = StatObject.mkTypeFcn(TyVar_list, tau)
	    val tystr = Environments.mkTyStr(typeFcn, Environments.emptyCE)

	    val (TE, out_typbind_opt) = elab_typbind_opt(C, typbind_opt)
	  in
	    (Environments.TE_plus_TE(Environments.singleTE(tycon, tystr), TE),
	     OG.TYPBIND(okConv i, tyvar_list, tycon, out_ty, out_typbind_opt))
	  end

     (*  Extended ML: "Syntax, Semantics and Dynamics Semantics of 
         Extended ML (version 2)" *)

       (* undefined typbind - rule 28.5 *)

     | IG.UNDEFtypbind(i,tyvars,tycon,typbind_opt) =>
         let

           (* Every time "?" is elaborated we choose a new type name *)

           val tau = freshType() 

           val TyVar_list =
	     List.map (fn tv => StatObject.mkExplicitTyVar tv) tyvars

           val typefcn = StatObject.mkTypeFcn(TyVar_list, tau)

           val tystr = Environments.mkTyStr(typefcn, Environments.emptyCE)
         
           val (TE, out_typbind_opt) = elab_typbind_opt(C, typbind_opt)
	 in
	   (Environments.TE_plus_TE(Environments.singleTE(tycon, tystr), TE),
	    OG.UNDEFtypbind(okConv i, tyvars, tycon, out_typbind_opt))
	 end

    and elab_typbind_opt (C : Environments.Context, typbind_opt : IG.typbind Option) :
      (Environments.TyEnv * OG.typbind Option) =

      case typbind_opt of

	Some(typbind) =>
	  let
	    val (TE, out_typbind) = elab_typbind(C, typbind)
	  in
	    (TE, Some out_typbind)
	  end

      | None =>
	  (Environments.emptyTE, None)

    (******* datatype bindings - Definition v3 page 27 *******)

    and elab_datbind (C : Environments.Context, datbind : IG.datbind)
      : ((Environments.VarEnv * Environments.TyEnv) * OG.datbind) =

      case datbind of

	(* Datatype binding *)
	IG.DATBIND(i, tyvar_list, tycon, conbind, datbind_opt) =>
	  let
	    val TyVar_list =
	      map (fn tv => StatObject.mkExplicitTyVar tv) tyvar_list

	    val (typeFcn, _) = 
	      case Environments.Lookup_tycon(C, tycon) of
		Some(tystr) => Environments.unTyStr(tystr)
	      | None => Crash.impossible "ElabDec.datbind(1)"

	    val tyname =
	      case StatObject.unTyName_TypeFcn(typeFcn) of
		Some(tyname) => tyname
	      | None => Crash.impossible "ElabDec.datbind(2)"

	    val tau_list =
	      map StatObject.mkTypeTyVar TyVar_list
	    val tau =
	      StatObject.mkTypeConsType(StatObject.mkConsType(tau_list, tyname))

(***
	    val cons = allCons conbind
 ***)
	    val (CE, out_conbind) = elab_conbind(C, tau, conbind)

	    val ClosCE = Environments.ClosCE CE
	    val tystr = Environments.mkTyStr(StatObject.TyName_in_TypeFcn tyname, ClosCE)

	    val ((VE, TE), out_datbind_opt) = elab_datbind_opt(C, datbind_opt)
	  in
	    ((Environments.VE_plus_VE(Environments.ClosCE_to_VE CE, VE),
	      Environments.TE_plus_TE(Environments.singleTE(tycon, tystr), TE)),
	      OG.DATBIND(okConv i, tyvar_list, tycon, out_conbind, out_datbind_opt))
	  end

    and elab_datbind_opt (C : Environments.Context, datbind_opt : IG.datbind Option)
      : ((Environments.VarEnv * Environments.TyEnv) * OG.datbind Option) =

      case datbind_opt of

	Some(datbind) =>
	  let
	    val ((VE, TE), out_datbind) = elab_datbind(C, datbind)
	  in
	    ((VE, TE), Some out_datbind)
	  end

       | None =>
	  ((Environments.emptyVE, Environments.emptyTE), None)

    (****** constructor bindings - Definition v4, 30 *****)

    and elab_conbind(C: Environments.Context,
		     tau: StatObject.Type,
		     conbind: IG.conbind
		    ): (Environments.ConEnv * OG.conbind) =

      case conbind of

	(* Constructor binding *)
	IG.CONBIND(i, IG.OP_OPT(con, withOp), Some ty, conbind_opt) =>
	  let
	    val (tau', out_ty) = elab_ty(C, ty)
	    val arrow =
	      StatObject.Type_in_TypeScheme(StatObject.mkTypeArrow(tau', tau))

	    val (CE, out_conbind_opt) =
	      elab_conbind_opt(C, tau, conbind_opt)
	  in
	    (Environments.CE_plus_CE(
	       Environments.singleCE(con, arrow), CE
	     ),
	     OG.CONBIND(okConv i, OG.OP_OPT(con, withOp),
			   Some out_ty, out_conbind_opt
		       )
	    )
	  end

	(* Constructor binding *)
      | IG.CONBIND(i, IG.OP_OPT(con, withOp), None, conbind_opt) =>
	  let
	    val (CE, out_conbind_opt) =
	      elab_conbind_opt(C, tau, conbind_opt)
	    val ts = StatObject.Type_in_TypeScheme tau
	  in
	    (Environments.CE_plus_CE(Environments.singleCE(con, ts), CE),
	     OG.CONBIND(okConv i, OG.OP_OPT(con, withOp), None, out_conbind_opt))
	  end

    and elab_conbind_opt(C: Environments.Context,
			 tau: StatObject.Type,
			 conbind_opt: IG.conbind Option
			): (Environments.ConEnv * OG.conbind Option) =

      case conbind_opt of

	Some(conbind) =>
	  let
	    val (CE, out_conbind) = elab_conbind(C, tau, conbind)
	  in
	    (CE, Some out_conbind)
	  end

      | None =>
	  (Environments.emptyCE, None)

    (****** exception bindings - Definition v4, 31..32 *****)

    and elab_exbind (C : Environments.Context, exbind : IG.exbind)
      : (Environments.ExConEnv * OG.exbind) =

      case exbind of

	(* Exception binding *)
	IG.EXBIND(i, IG.OP_OPT(excon, withOp), Some ty, rest) =>
	  let
	    val (tau, out_ty) = elab_ty(C, ty)

  (*XX UNFINISHED*)

	    (* UNIMPLEMENTED: check that tau is imperative. *)

	    val exnTy = StatObject.mkTypeArrow(tau, StatObject.TypeExn)
	    val EE_this = Environments.singleEE(excon, exnTy)
	    val (EE_rest, out_rest) = elab_exbind_opt(C, rest)
	  in
	    (Environments.EE_plus_EE(EE_this, EE_rest),
	     OG.EXBIND(okConv i, OG.OP_OPT(excon, withOp), Some out_ty, out_rest))
	  end

	(* Exception binding *)
      | IG.EXBIND(i, IG.OP_OPT(excon, withOp), None, rest) =>
	  let
	    val EE_this = Environments.singleEE(excon, StatObject.TypeExn)
	    val (EE_rest, out_rest) = elab_exbind_opt(C, rest)
	  in
	    (Environments.EE_plus_EE(EE_this, EE_rest),
	     OG.EXBIND(okConv i, OG.OP_OPT(excon, withOp), None, out_rest))
	  end

	(* Exception binding *)
      | IG.EXEQUAL(i, IG.OP_OPT(excon, exconOp),
		      IG.OP_OPT(longid, longidOp), rest
		  ) =>
	  (case Environments.Lookup_longid(C, longid) of

	     Some(Environments.LONGEXCON tau) =>
	       let
		 val EE_this = Environments.singleEE(excon, tau)
		 val (EE_rest, out_rest) = elab_exbind_opt(C, rest)
		 val longexcon = R.LONGEXCON(StatObject.mk_longexcon longid)
	       in
		 (Environments.EE_plus_EE(EE_this, EE_rest),
		  OG.EXEQUAL(okConv i, OG.OP_OPT(excon, exconOp),
				OG.OP_OPT(longexcon, longidOp), out_rest
			    )
		 )
	       end

	   | _ =>
	       let
		 (* Carry on, building an error node. *)
		 val (EE_rest, out_rest) = elab_exbind_opt(C, rest)
		 val longexcon = R.LONGEXCON(StatObject.mk_longexcon longid)
	       in
		 (EE_rest, OG.EXEQUAL(
			     lookupIdError(i, longid),
			     OG.OP_OPT(excon, exconOp),
			     OG.OP_OPT(longexcon, longidOp),
			     out_rest
			   )
		 )
	       end)

    and elab_exbind_opt(C, Some exbind) =
	  let
	    val (EE, out_exbind) = elab_exbind(C, exbind)
	  in
	    (EE, Some out_exbind)
	  end

      | elab_exbind_opt(C, None) =
	  (Environments.emptyEE, None)

    (****** atomic patterns - Definition v3 page 28   ******)

    and elab_atpat (C : Environments.Context, atpat : IG.atpat) :
	((Environments.VarEnv * StatObject.Type) * OG.atpat) =

	case atpat of

	  (* Wildcard *)
	  IG.WILDCARDatpat i  =>
	    ((Environments.emptyVE, freshType()),
	      OG.WILDCARDatpat(okConv i))

	  (* Special constant *)
	| IG.SCONatpat(i,scon) =>
	    ((Environments.emptyVE, StatObject.GetTypescon scon),
	      OG.SCONatpat(okConv i,scon))

	  (* Long identifier *)
	| IG.LONGIDatpat(i, IG.OP_OPT(longid, withOp)) =>
	    (case Environments.Lookup_longid(C, longid)
	       of Some(Environments.LONGCON sigma) =>
		    let
		      fun isConsType tau =
			case StatObject.unTypeConsType tau
			  of Some _ => true
			   | None => false

		      val tau = StatObject.instance sigma
		      
		      val (tau', i') =
			if isConsType tau then
			  (tau, okConv i)
			else
			  (bogus_Type(),
			   errorConv(i, ErrorInfo.NOTCONSTYPE tau)
			  )
		    in
		      ((Environments.emptyVE, tau'),
		       let
			 val longcon = StatObject.mk_longcon longid
		       in
			 OG.LONGIDatpat(
			   addTypeInfo_CON(i', C,
					   StatObject.isTypeArrow tau', longid
					  ),
			   OG.OP_OPT(R.LONGCON longcon, withOp)
			)
		       end
		      )
		    end

	        | Some(Environments.LONGEXCON tau) =>
		    let
		      val longexcon = StatObject.mk_longexcon longid
		      val exnType = StatObject.TypeExn
		      val (_, i') = Unify(tau, exnType, i)
		    in
		      ((Environments.emptyVE, exnType),
		       OG.LONGIDatpat(i',
				      OG.OP_OPT(R.LONGEXCON longexcon, withOp)
				     )
		      )
		    end

		| _ =>		(* make new variable environment *)
		    let
		      val var = StatObject.mk_longvar longid
		      val tau = freshType()
		      val tau_scheme = StatObject.Type_in_TypeScheme tau
		    in
		      case StatObject.decomposeLongId longid
			of (nil, id) =>
			     ((Environments.singleVarVE(id, tau_scheme), tau),
			      OG.LONGIDatpat(okConv i,
					     OG.OP_OPT(R.LONGVAR var, withOp)
					    )
			     )

		         | (_, _) =>
			     ((Environments.bogus_VE, bogus_Type()),
			      OG.LONGIDatpat(
				errorConv(i, ErrorInfo.QUALIFIED_ID longid),
				OG.OP_OPT(R.LONGVAR var, withOp)
			      )
			     )
		    end
	    )

	  (* Record pattern *)
	| IG.RECORDatpat(i, row_opt as None) =>
	    ((Environments.emptyVE, StatObject.TypeUnit),
	      OG.RECORDatpat(okConv i, None)
	     )

	  (* Record pattern *)
	| IG.RECORDatpat(i, row_opt as Some patrow) =>
	    let
	      val ((VE, rho), out_patrow) = elab_patrow(C, patrow)
	    in
	      ((VE,StatObject.mkTypeRecType rho),
		OG.RECORDatpat(okConv i, Some(addLabelInfo(rho, out_patrow)))
	      ) 
	    end

	  (* Parenthesised pattern *)
	| IG.PARatpat(i, pat) =>
	    let val ((VE,tau), out_pat) = elab_pat(C, pat)
	    in ((VE,tau), OG.PARatpat(okConv i,out_pat)) end

    (****** pattern rows - Definition v3 page 28 ******)

    and elab_patrow(C: Environments.Context, patrow: IG.patrow)
	  : ((Environments.VarEnv * StatObject.RecType) * OG.patrow) =
      case patrow
	of IG.PATROW(i, lab, pat, None) =>
	     let
	       val ((VE, tau), out_pat) = elab_pat(C, pat)
	     in
	       ((VE, StatObject.addField (lab, tau) StatObject.emptyRecType),
		OG.PATROW(okConv i, lab, out_pat, None)
	       )
	     end

	 | IG.PATROW(i, lab, pat, Some patrow) =>
	     let
	       val ((VE, tau), out_pat) = elab_pat(C, pat)
	       val ((VE', rho), out_patrow) = elab_patrow(C, patrow)
	     in
	       ((Environments.VE_plus_VE(VE, VE'),
		 StatObject.addField (lab, tau) rho
		), OG.PATROW(okConv i, lab, out_pat, Some out_patrow)
	       )
	     end

	| IG.DOTDOTDOT i =>		(* Flexible record treatment... *)
	    ((Environments.emptyVE, StatObject.emptyFlexRecType()),
	     OG.DOTDOTDOT(okConv i)
	    )

    (****** patterns - Definition v3 page 29 ******)

    and elab_pat (C : Environments.Context, pat : IG.pat) :
	((Environments.VarEnv * StatObject.Type) * OG.pat) =
      let
	val _ =
	  if Flags.DEBUG_ELABDEC then
	    pr("elab_pat: ", PPInDecGrammar.layoutPat pat)
	  else ()

	val ((VE, ty), pat') = elab_pat'(C, pat)

	val _ =
	  if Flags.DEBUG_ELABDEC then
	    let 
	      val t = PP.NODE{start="{", finish="}", indent=0,
			      children=[Environments.layoutVE VE,
					StatObject.layoutType ty
				       ],
			      childsep=PP.RIGHT "; "
			     }
	    in
	      pr("giving:   ", t)
	    end
	      else ()
      in
	((VE, ty), pat')
      end

    and elab_pat'(C, pat) =
	case pat of

	  (* Atomic pattern *)
	  IG.ATPATpat(i, atpat) =>
	    let val ((VE,tau), out_atpat) = elab_atpat(C, atpat)
	    in ((VE,tau), OG.ATPATpat(okConv i,out_atpat)) end

	  (* Constructed pattern *)
	| IG.CONSpat(i, IG.OP_OPT(longid, withOp), atpat) =>
	    let
	      val ((VE,tau'), out_atpat) = elab_atpat(C, atpat)
	    in
	      case Environments.Lookup_longid(C, longid) of

		Some(Environments.LONGCON sigma) =>
		  let
		    val new = freshType()
		    val arrow = StatObject.mkTypeArrow(tau', new) 
		    val tau1 = StatObject.instance sigma
		    val (S1, i') = Unify(arrow, tau1, i)
		    val longcon = StatObject.mk_longcon longid
		  in
		    ((S1 onVE VE, S1 on new),
		     OG.CONSpat(addTypeInfo_CON(i', C, true, longid),
				OG.OP_OPT(R.LONGCON longcon, withOp),
				out_atpat
			       )
		    )
		  end

	      | Some(Environments.LONGEXCON tau) =>
		  let
		    val arrow = StatObject.mkTypeArrow(tau',StatObject.TypeExn)
		    val (S1, i') = Unify(arrow, tau, i)
		    val longexcon = StatObject.mk_longexcon longid
		  in
		    ((S1 onVE VE,StatObject.TypeExn),
		     OG.CONSpat(i', OG.OP_OPT(R.LONGEXCON longexcon, withOp),
				    out_atpat
			       )
		    )
		  end

(***KEVIN's idea for expressions as patterns:

	      | Some(Environments.LONGVAR sigma) =>
		  let
		    val new = freshType()
		    val arrow = StatObject.mkTypeArrow(new, tau')
		    val tau = StatObject.instance sigma
		    val (S1, i') = Unify(arrow,tau,i)
		    val longvar = StatObject.mk_longvar longid
		  in
		    ((S1 onVE VE, S1 on new),
		     OG.CONSpat(i', OG.OP_OPT(R.LONGVAR longvar, withOp),
				    out_atpat
			       )
		    )
		  end

***)

	      | _ => (* Mark the error. *)
		  ((VE, bogus_Type()),
		   OG.CONSpat(lookupIdError(i, longid),
			      OG.OP_OPT(R.LONGCON StatObject.bogusCon, withOp),
			      out_atpat
			     )
		  )
	    end

	  (* Typed pattern *)
	| IG.TYPEDpat(i, pat, ty) =>
	    let
	      val ((VE,tau), out_pat) = elab_pat(C, pat)
	      val (tau', out_ty) = elab_ty(C, ty)
	      val (S, i') = Unify(tau, tau', i)
	    in
	      ((S onVE VE, S on tau), OG.TYPEDpat(i', out_pat, out_ty))
	    end

	| IG.LAYEREDpat(i, IG.OP_OPT(id, withOp), None, pat) =>
	    let
	      val ((VE1, tau), out_pat) = elab_pat(C, pat)
	      val VE2 = Environments.singleVarVE(id, StatObject.Type_in_TypeScheme tau)
	      val VE3 = Environments.VE_plus_VE(VE1, VE2)
	      val var = StatObject.mk_var id
	    in
	      ((VE3, tau), OG.LAYEREDpat(okConv i, OG.OP_OPT(var, withOp),
					    None, out_pat
					)
	      )
	    end

	| IG.LAYEREDpat(i, IG.OP_OPT(id, withOp), Some ty, pat) =>
	    let
	      val ((VE1, tau), out_pat) = elab_pat(C, pat)
	      val (tau', out_ty) = elab_ty(C, ty)
	      val (S, i') = Unify(tau, tau', i)

	      val VE2 = Environments.singleVarVE(id, StatObject.Type_in_TypeScheme tau)
	      val var = StatObject.mk_var id
	      val VE3 = Environments.VE_plus_VE(VE1, VE2)
	    in
	      ((S onVE VE3, S on tau),
	       OG.LAYEREDpat(i', OG.OP_OPT(var, withOp), Some out_ty, out_pat)
	      )
	    end

	| IG.UNRES_INFIXpat _ =>
	    Crash.impossible "elab_pat(UNRES_INFIX)"

    (****** types - Definition v3 page 29 ******)

    and elab_ty (C :  Environments.Context, ty : IG.ty) :
	(StatObject.Type * OG.ty) =

	case ty of

	  (* Explicit type variable *)
	  IG.TYVARty(i, tyvar) =>
	    (StatObject.mkTypeTyVar(StatObject.mkExplicitTyVar tyvar),
	     OG.TYVARty(okConv i,tyvar))

	  (* Record type *)
	| IG.RECORDty(i, None) =>
	    (StatObject.TypeUnit, OG.RECORDty(okConv i,None))

	  (* Record type *)
	| IG.RECORDty(i, Some tyrow) =>
	    let
	      val (rho, out_tyrow) = elab_tyrow(C, tyrow)
	    in
	      (StatObject.mkTypeRecType rho,
	       OG.RECORDty(okConv i, Some out_tyrow))
	    end

	  (* Constructed type *)
	| IG.CONty(i, ty_list, longtycon) =>
	    let
	      val res_list    = map (fn ty => elab_ty (C, ty)) ty_list
	      val tau_list    = map #1 res_list
	      val out_ty_list = map #2 res_list
	    in
	      case Environments.Lookup_longtycon(C, longtycon)
		of Some tystr =>
		     let
		       val (typeFcn, _) = Environments.unTyStr tystr
		       val expectedArity = StatObject.arity_TypeFcn typeFcn
		       val actualArity = List.size tau_list
		     in
		       if expectedArity = actualArity then
			 (StatObject.applyTypeFcn(typeFcn, tau_list),
			  OG.CONty(okConv i, out_ty_list, longtycon)
			 )
		       else
			 (bogus_Type(),
			  OG.CONty(errorConv(i, ErrorInfo.WRONG_ARITY{
						  expected=expectedArity,
						  actual=actualArity
						}
					    ),
				   out_ty_list, longtycon
				  )
			 )
		     end

	         | None =>
		       (bogus_Type(),
			OG.CONty(
			  lookupTyConError(i, longtycon),
			  out_ty_list, longtycon
			)
		       )
	    end

	  (* Function type *)
	| IG.FNty(i, ty, ty') =>
	    let
	      val (tau , out_ty ) = elab_ty(C, ty )
	      val (tau', out_ty') = elab_ty(C, ty')
	    in
	      (StatObject.mkTypeArrow(tau, tau'),
	       OG.FNty(okConv i, out_ty, out_ty'))
	    end

	  (* Parenthesised type *)
	| IG.PARty(i, ty) =>
	    let
	      val (tau, out_ty) = elab_ty(C, ty)
	    in
	      (tau, OG.PARty(okConv i, out_ty))
	    end

    (****** type rows - Definition v3 page 30 ******)

    and elab_tyrow (C :  Environments.Context, tyrow : IG.tyrow) :
	(StatObject.RecType * OG.tyrow) =

	case tyrow of

	  (* Type row *)
	  IG.TYROW(i, lab, ty, None) =>
	    let
	      val (tau, out_ty) = elab_ty(C, ty)
	      val rho = StatObject.addField (lab,tau) StatObject.emptyRecType
	    in
	      (rho, OG.TYROW(okConv i, lab, out_ty, None))
	    end

	  (* Type row *)
	| IG.TYROW(i, lab, ty, Some tyrow) =>
	    let
	      val (tau, out_ty) = elab_ty(C, ty)
	      val (rho, out_tyrow) = elab_tyrow(C, tyrow)
	    in
	      (StatObject.addField (lab,tau) rho,
	       OG.TYROW(okConv i, lab, out_ty, Some out_tyrow))
	    end


    (*  Extended ML: "Syntax, Semantics and Dynamics Semantics of 
        Extended ML (version 2)" *)

    (* Axioms - rule 32.5 *)

    and elab_axiom(C : Environments.Context, axiom : IG.ax as IG.AXIOMS(i, exp, ax_opt)) :
                  (Environments.Env * OG.ax) =
       let
         fun elab_ax_opt(C, axiom) =
             case axiom of
                  None     => None
                | Some(ax) => 
                     let 
                       val (E,out_ax) = elab_axiom(C, ax)
                     in 
                       Some(out_ax) 
                     end

        (*  We expect the resulting environment E to be empty but 
            elaborate anyway for typechecking -
            first construct the context according to rule
            32.5  *) 

       val VE = Environments.equalsVE(Environments.TE_of_C(C))
       val C' = C C_cplus_E(Environments.VE_in_E VE)       

       val (S, tau, out_exp) = elab_exp(C', exp)
       val  boolean = is_bool(Environments.TE_of_C(C'),tau)
     in
       if boolean then
          (Environments.emptyE, OG.AXIOMS(okConv i, out_exp, elab_ax_opt(C, ax_opt)))
       else 
          let
            val i' = logexp_error(okConv i,tau)
          in
            (Environments.emptyE, OG.AXIOMS(i', out_exp, elab_ax_opt(C, ax_opt)))
          end
     end


    (****** Elaborate a declaration ******)

    val elab_dec : (Environments.Context * IG.dec) ->
      (Environments.Env * OG.dec) =

      fn (C, dec) =>
	let
	  val (_, E, out_dec) = elab_dec(C, dec)
	in
	  (E, out_dec)
	end
  end;
