(*
$File: Common/Prelude.sml $
$Date: 1992/04/29 16:37:37 $
$Revision: 1.10 $
$Locker:  $
*)

(* Prelude for the Kit. *)

infix  3 := o
infix  4 = <> < #< > #> <= #<= >= #>=
infix  5 @
infixr 5 ::
infix  6 + #+ - #- ^
infix  7 div mod / * #*

exception Abs
      and Ord
      and Chr
      and Div
      and Mod
      and Quot
      and Prod
      and Neg
      and Sum
      and Diff
      and Floor
      and Sqrt
      and Exp
      and Ln
      and Io of string
      and Match
      and Bind
      and Interrupt

type unit = { }					
datatype bool = true | false

val not = fn true => false | false => true

datatype 'a list = nil | op :: of 'a * 'a list

fun op = (x: ''a, y: ''a): bool = prim(0, (x, y))
    and abs(x: int): int = prim(1, (x, Abs))
    and floor(x: real): int = prim(2, (x, Floor))
    and real(x: int): real = prim(3, x)
    and sqrt(x: real): real = prim(4, (x, Sqrt))
    and sin(x: real): real = prim(5, x)
    and cos(x: real): real = prim(6, x)
    and arctan(x: real): real = prim(7, x)
    and exp(x: real): real = prim(8, (x, Exp))
    and ln(x: real): real = prim(9, (x, Ln))
    and size(x: string): int = prim(10, x)
    and chr(x: int): string = prim(11, (x, Chr))
    and ord(x: string): int = prim(12, (x, Ord))
    and explode(x: string): string list = prim(13, x)
    and implode(x: string list): string = prim(14, x)
    and (x: int) + (y: int): int = prim(15, (x, y, Sum))
    and (x: int) - (y: int): int = prim(16, (x, y, Diff))
    and (x: int) * (y: int): int = prim(17, (x, y, Prod))
    and (x: real) / (y: real): real = prim(18, (x, y, Quot))
    and (x: int) div (y: int): int = prim(19, (x, y, Div))
    and (x: int) mod (y: int): int = prim(20, (x, y, Mod))
    and (x: int) < (y: int): bool = prim(21, (x, y))
    and (x: int) > (y: int): bool = prim(22, (x, y))
    and ~(x: int): int = prim(23, (x, Neg))
    and (x: 'a ref) := (y: 'a): unit = prim(24, (x, y))
    and !(x: 'a ref): 'a = prim(25, x)

	(* `!' is primitive to make the type non-weak (the Definition
	   is inconsistent here).
	*)

   (* Because we don't have overloading yet, we'll provide monotyped
      operations on reals. *)

    and abs_R(x: real): real = prim(26, (x, Abs))
    and #~(x: real): real = prim(27, (x, Neg))
    and (x: real) #+ (y: real): real = prim(28, (x, y, Sum))
    and (x: real) #- (y: real): real = prim(29, (x, y, Diff))
    and (x: real) #* (y: real): real = prim(30, (x, y, Prod))
    and (x: real) #< (y: real): bool = prim(31, (x, y))
    and (x: real) #> (y: real): bool = prim(32, (x, y))

fun (f o g) x = f(g x)

val op <> = not o (op =)
val op <= = not o (op >)
val op >= = not o (op <)
val op #>= = not o (op #<)
val op #<= = not o (op #>)

fun nil @ M = M
  | (x :: L) @ M = x :: (L @ M)

fun map F nil = nil
  | map F (x :: L) = F x :: map F L

fun rev nil = nil
  | rev (x :: L) = rev L @ [x]

fun s ^ s' = implode(explode s @ explode s')

(* Our I/O streams are just numbers, so that we don't have to implement
   any primitive type in the interpreter. *)

abstype instream = INSTREAM of int
    and outstream = OUTSTREAM of int
with
  val std_in = INSTREAM 0
  val std_out = OUTSTREAM 0

  fun open_in(f: string) = INSTREAM(prim(33, (f, Io("Cannot open " ^ f))))

  fun open_out(f: string) = OUTSTREAM(prim(34, (f, Io("Cannot open " ^ f))))

  fun input(INSTREAM i, n: int): string = prim(35, (i, n))
    
  fun lookahead(INSTREAM i): string = prim(36, i)

  fun close_in(INSTREAM i): unit = prim(37, i)

  fun end_of_stream(INSTREAM i): bool = prim(38, i)
    
  fun output(OUTSTREAM i, str: string): unit =
    prim(39, (i, str, Io "Output stream is closed"))

  fun close_out(OUTSTREAM i): unit = prim(40, i)
end;
  (*
structure Kit : sig
                  exception Use
                  val use : string -> unit
                end =
  struct
    exception Use
    fun use(f): unit = prim(41, (f, Io f, Use))
  end;


*)

