(*
$File: Compiler/COMPILER_ENV.sml $
$Date: 1991/12/03 15:54:35 $
$Revision: 1.1 $
$Locker:  $
*)

(*$COMPILER_ENV*)
(* COMPILER_ENV: used by CompileDec to look up identifiers, returning lvars
   (for things which have actually been declared) or prims (for built-in
   functions). *)

signature COMPILER_ENV =
  sig
    type CEnv
    type var				(* Unqualified variables. *)
    type longvar			(* Qualified vars (for lookup). *)
    type excon				(* Unqualified exception constructors.*)
    type longexcon			(* Qualified excons (for lookup). *)
    type lvar				(* Unique lambda identifiers. *)

    val emptyCEnv: CEnv
    val initialCEnv: CEnv

    val declareVar: (var * lvar * CEnv) -> CEnv
    val declareExcon: (excon * lvar * CEnv) -> CEnv

    val plus: CEnv * CEnv -> CEnv

    datatype result = LVAR of lvar
                    | PRIM		(* Looking up a value identifier will
					   either get you a real lvar,
					   or PRIM (which means you have to
					   extract the integer argument and
					   use it in a PRIM lambda-exp). *)

    val lookupLongvar: CEnv -> longvar -> result
					(* Only `prim' should give you back
					   `PRIM' (and even then it can be
					   overwritten). *)

    val lookupLongexcon: CEnv -> longexcon -> lvar
					(* Looking up an excon must give
					   you an lvar. *)

    val lookupExcon: CEnv -> excon -> lvar
					(* MEMO: Temporary: Match and Bind. *)

    val lvarsOfCEnv: CEnv -> lvar list

    type StringTree
    val layoutCEnv: CEnv -> StringTree
  end;
