(* Dynamic values and enviroments - Definition v3 page 46 *)

(*
$File: Interpreter/DYNOBJECT.sml $
$Date: 1991/12/03 15:56:07 $
$Revision: 1.1 $
$Locker:  $
*)

(*$DYNOBJECT: CON LAB VAR EXCON EXNAME*)

signature DYNOBJECT =
  sig
    structure Con: CON
    structure Lab: LAB
    structure Var: VAR
    structure Excon: EXCON
    structure ExName: EXNAME
    structure Grammar: sig  type match  end

   (* Special values (constants): *)

    eqtype scon and SVal
    val mkSValSCon: scon -> SVal

   (* We choose to keep the environments abstract: *)

    type Env
    type VarEnv
    type ExConEnv
    type Closure

   (* Values (now concrete): *)

    type BasVal
    type ('a, 'b) map

    datatype Val = Sval       of SVal
		 | BASval     of BasVal
		 | CON0val    of Con.con
		 | CON1val    of Con.con * Val
		 | RECORDval  of (Lab.lab, Val) map
		 | EXNAME0val of ExName.ExName
		 | EXNAME1val of ExName.ExName * Val
		 | CLOSUREval of Closure
                 | INCOMPLETEval

   (* Variable environments *)

    val emptyVE: VarEnv
    and singleVE: Var.var * Val -> VarEnv
    and plusVE: VarEnv * VarEnv -> VarEnv

   (* Exception environments *)

    val emptyEE: ExConEnv
    and singleEE: Excon.excon * ExName.ExName -> ExConEnv
    and plusEE: ExConEnv * ExConEnv -> ExConEnv
    and lookupEE_excon: ExConEnv -> Excon.excon -> ExName.ExName

   (* Environments *)

    val emptyE: Env
    and plusE: Env * Env -> Env
    and VE_in_E: VarEnv -> Env
    and EE_in_E: ExConEnv -> Env
    and EE_of_E: Env -> ExConEnv

    val initialE: Env

   (* Closures *)

    val mkClosure: Grammar.match * Env * VarEnv -> Closure
    and unClosure: Closure -> Grammar.match * Env * VarEnv

   (* Other operations *)

    val Rec : VarEnv -> VarEnv	(* for the unfolding of recursive closures *)

   (* Lookup functions *)

    val lookupEnv_longvar: Env -> Var.longvar -> Val
    val lookupEnv_longexcon: Env -> Excon.longexcon -> ExName.ExName

   (* Exception packets
      Includes a single exception to model all exceptions.
      Needs to be here, since APPLY can raise exceptions from
      the pervasives. *)

    datatype Pack = EXNAME0pack of ExName.ExName
		  | EXNAME1pack of ExName.ExName * Val

    exception EXCEPTION of Pack

   (* Printing *)

    type StringTree
    val layoutVal: Val -> StringTree
    val layoutEnv: Env -> StringTree
  end;






