(* Dynamic values and enviroments - Definition v3 page 46 *)

(*
$File: Interpreter/DynObject.sml $
$Date: 1991/12/03 15:56:07 $
$Revision: 1.1 $
$Locker:  $
*)

(*$DynObject:
	FINMAP LAB CON SPECIAL_VALUE VAR EXCON EXNAME BASIC_VALUE
	PRETTYPRINT CRASH DYNOBJECT
 *)

functor DynObject(structure FinMap: FINMAP
		  structure Lab: LAB
		  structure Con: CON
		  structure SpecialValue: SPECIAL_VALUE
		  structure Var: VAR
		  structure Excon: EXCON
		  structure ExName: EXNAME
		  structure Grammar: sig type match end
		  structure BasicValue: BASIC_VALUE

		  structure PP: PRETTYPRINT
		    sharing type PP.StringTree = FinMap.StringTree

		  structure Crash: CRASH
		 ): DYNOBJECT =
  struct
   (* Copying of arguments to result structure *)

    structure Con     = Con
    structure Lab     = Lab
    structure Var     = Var
    structure Excon   = Excon
    structure ExName  = ExName
    structure Grammar = Grammar

    type SVal   = SpecialValue.SVal
    type scon   = SpecialValue.scon
    type BasVal = BasicValue.BasVal

    type ('a, 'b) map = ('a, 'b) FinMap.map
    type StringTree = PP.StringTree


    datatype Val = Sval       of SVal
		 | BASval     of BasVal
		 | CON0val    of Con.con
		 | CON1val    of Con.con * Val
		 | RECORDval  of (Lab.lab, Val) FinMap.map
		 | EXNAME0val of ExName.ExName
		 | EXNAME1val of ExName.ExName * Val
		 | CLOSUREval of Closure
                 | INCOMPLETEval

   (* These are abstract to the outside world: *)
    and Closure  = CLOSURE  of Grammar.match * Env * VarEnv
    and Env      = ENV      of VarEnv * ExConEnv
    and VarEnv   = VARENV   of (Var.var, Val) FinMap.map
    and ExConEnv = EXCONENV of (Excon.excon, ExName.ExName) FinMap.map


    fun mkSValSCon sc = SpecialValue.mkSValSCon sc

    val mkClosure = CLOSURE
    fun unClosure(CLOSURE c) = c

   (* for environments *)
   (* MEMO: not full ML; structure environments not implemented *)

   (* Empty enviroments *)

    val emptyVE = VARENV FinMap.empty
    val emptyEE = EXCONENV FinMap.empty

    (* for variable environments *)

    fun singleVE(v, value) = VARENV(FinMap.singleton(v, value))
    fun singleEE(e, name)  = EXCONENV(FinMap.singleton(e, name))

    val initialVE = singleVE(Var.var_PRIM, BASval BasicValue.prim)

    val emptyE = ENV(emptyVE, emptyEE)

    val initialE = ENV(initialVE, emptyEE)

    fun VE_in_E VE = ENV(VE, emptyEE)
    fun EE_in_E EE = ENV(emptyVE, EE)
    fun EE_of_E (ENV(_, EE)) = EE

    (* Unfolding of recursive closures *)

    fun Rec(VE as VARENV m) =
      let
	fun unfold(CLOSUREval(CLOSURE(match, E', VE''))) =
	      CLOSUREval(CLOSURE(match, E', VE))
	  | unfold v = v
      in
	VARENV(FinMap.composemap unfold m)
      end

    (******** Lookup functions ********)

    fun lookupEnv_longvar(ENV(VARENV m, _)) longvar =
      case (Var.decompose longvar) of
	([], var) =>
	   (case FinMap.lookup m var
	      of Some value  => value
	       | None =>
		   Crash.impossible "lookupEnv_var"
	      )
      | (_ , var) =>
	   Crash.unimplemented "DynObject.lookupEnv_longvar"

    fun lookupEnv_longexcon(ENV(_, EXCONENV m)) longexcon =
      case (Excon.decompose longexcon) of
	([], excon) =>
	   (case (FinMap.lookup m excon) of
	      Some(name) => name
	    | None => Crash.impossible "lookupEnv_excon")
      | (_ ,excon) =>
	   Crash.unimplemented "DynObject.lookupEnv_longexcon"

    fun lookupEE_excon(EXCONENV m) excon =
      case (FinMap.lookup m excon) of
	Some(name) => name
      | None => Crash.impossible "lookupEnv_excon"

    fun plusE(ENV(VARENV a, EXCONENV b),ENV(VARENV c, EXCONENV d)) =
      ENV(VARENV(FinMap.plus(a, c)), EXCONENV(FinMap.plus(b, d)))

    fun plusVE(VARENV m, VARENV m') = VARENV(FinMap.plus(m, m'))
    fun plusEE(EXCONENV m, EXCONENV m') = EXCONENV(FinMap.plus(m, m'))

   (* exception packets. *)

    datatype Pack = EXNAME0pack of ExName.ExName
		  | EXNAME1pack of ExName.ExName * Val

    exception EXCEPTION of Pack

   (* printing. *)

    val layoutLab = PP.layoutAtom Lab.pr_lab
    val layoutVar = PP.layoutAtom Var.pr_var

    fun layoutVal(Sval sval) = PP.layoutAtom SpecialValue.pr_SVal sval
      | layoutVal(BASval bv) = PP.LEAF "fn"
      | layoutVal(CON0val c) = PP.layoutAtom Con.pr_con c

      | layoutVal(CON1val(c, v)) =
	  PP.NODE{start="(", finish=")", indent=1,
		  children=[PP.layoutAtom Con.pr_con c, layoutVal v],
		  childsep=PP.RIGHT ", "
		 }

      | layoutVal(CLOSUREval _) = PP.LEAF "fn"

      | layoutVal(EXNAME0val exname) =
	  PP.layoutAtom ExName.pr_ExName exname

      | layoutVal(EXNAME1val(exname, v)) =
	  PP.NODE{start=ExName.pr_ExName exname ^ "(", finish=")", indent=3,
		  children=[layoutVal v], childsep=PP.NONE
		 }

      | layoutVal(RECORDval m) =
	  FinMap.layoutMap {start="{", eq="=", sep=", ", finish="}"}
			   layoutLab layoutVal m

      | layoutVal INCOMPLETEval = PP.LEAF "Incomplete Value"

    fun layoutEnv(ENV(VARENV m, _)) =
      FinMap.layoutMap {start="{", eq=" = ", sep=", ", finish="}"}
		       layoutVar layoutVal m
  end;
